! -*-f90-*-
! Originator:cbrindle
! Adapted:   Jorgo Bakker (jbakker@rssd.esa.int)
! Info:
!   Interactive program for selecting sources from an image
!   for further analysis
! Note:
!  * Should always call "omsource_init()" first, and "omsource_finish()" last.
!  * In comments, I usually talk about methods. A method may be either a
!    function OR subroutine within a module
!  * Any variable starting with 'my' is global to this module
module omsource_old
   use dal
   use cal
   use sasdal
   use oal
   use param
   use sla
   use errorHandling
   

!   use frametime_mod, only : getFrameTime ! mssllib

   use omsourceparameter_mod
   use omsourcedetect_mod
   use omsourcephotometry_mod
   use omsourcecelestial_mod
   use omprogress_mod
   use omsourceregion_mod
   use omsourceutils_mod
   use omsourcebackground_mod
   
   implicit none


   private ! all, but:
   public  :: OmSourceParameter   ! type definition
   public  :: OmSourceRegion      ! type definition
   public :: ObsDetails           ! NB
   public :: ImageWindow          !

   public  :: omsource_init       ! subroutine ...()
   public  :: omsource_finish     ! subroutine ...()
   public  :: omsource_loadsources  ! subroutine ... ()
   public  :: omsource_delete     ! subroutine ...()
   public  :: omsource_recompute  ! subroutine ...()
   public  :: omsource_add        ! subroutine ...()


   interface construct
      module procedure construct_od
   end interface

   interface destruct
      module procedure destruct_od
   end interface
 
   interface construct
      module procedure construct_iw
   end interface

   interface destruct
      module procedure destruct_iw
   end interface
   !TODO>


   ! local storage types:
   type(ImageWindow)      :: myImageWindow
   type(ObsDetails)       :: myObsDetails
   type(OmCelestial)      :: myCelestial

   real(single),parameter :: myNullValue=99999.

   ! image related stuff:
   real(single), pointer  :: myImage(:, :)   ! spatial
   logical(bool),pointer  :: myQuality(:, :) ! quality array
   real(single)           :: myTrimage(1:6)  ! (shortcut)


   ! The aperture radii etc...
   real(single)           :: myApertureRadius
   real(single)           :: myBgrRadiusIn
   real(single)           :: myBgrRadiusOut
   real(single)           :: myBgrMode
!   real(single)           :: myExposure

   ! Some CCD parameters for computing magnitudes:
   real(double)           :: myDeadFraction
   real(double)           :: myFrameTime
   real(single)           :: myDetectionRate
   real(single)           :: myDetectionMagnitude


   ! datasets in memory:
   type(DataSetT)         :: myImageSet

   ! determine whether we loaded sources or not?
   ! TODO: this is used for myDetectionRate/Magnitude, but
   ! if the user starts adding sources, is this still valid?
   logical                :: mySourcesAreLoaded
   


contains

!*********************************************************************
! Info: Initialises the fortran part of omsource.
!       It loads the image pdata and properties, and -if any- it also
!       creates a source list from an already existing sources dataset
!*********************************************************************
subroutine omsource_init(sources, &
                          
                         photometricParameters, &
                         observationDetails, &
                         imageWindowDetails, imageFileName, oldSourceListFile)

      type(OmSourceParameter),  pointer    :: sources(:)
      type(PhotometryParameters), intent(out) :: photometricParameters
      type(ObsDetails), intent(out)  :: observationDetails
      type(ImageWindow), intent(out) ::  imageWindowDetails
      character(*), intent(in) :: imageFileName, oldSourceListFile
            
!      character(128) :: imageFileName, sourceListFileName
      type(OmProgressT)        :: reporter    ! ... of progress to GUI
      character(*) , parameter :: routine="omsource_old.omsource_init"
!      character(256)  oldSourceListFile
 

      nullify(sources)       ! proper state

      ! --- initialise globals in this module
!imageFileName = fileNames(1 : 128)
!sourceListFileName = fileNames(129 : 256)
!write(*, *) "omsourceold_mod.f90 lin 167 image file name=",  imageFileName
!write(*, *) "source list  file name=",  sourceListfileName

      myDeadFraction       = -1.
      myFrameTime          = -1.
      myDetectionRate      = -1.
      myDetectionMagnitude = -1.

      nullify(myImage)
      nullify(myQuality)

      ! --- loading general information ---
      call construct(reporter, label="Initialising - please wait", steps = 10)
      call setProgress(reporter, 1)

      ! --- extract the image from the file
!      myImageSet = dataSet(stringParameter("imageset"), read)
      myImageSet = dataSet(imageFileName, read)


      call setProgress(reporter,2)

      ! load the image parameters (WINDOWX, etc) from the image fits file...
      call construct(myImageWindow, set = myImageSet)
      call setProgress(reporter,3)

      ! set the frame time etc...


      myFrameTime = myImageWindow%frameTime
      myDeadFraction = myImageWindow%deadFraction


      call setProgress(reporter,4)

      !<TODO: missing DATATYPE in dataset, otherwise we could do this:
      !call CAL_setState(myImageSet)
      !call setProgress(reporter,5)
      !TODO>

      ! create celestial information; also sets state of CAL
      call construct(myCelestial, set=myImageSet)
      call setProgress(reporter, 6)

      ! now init radii once, and once only, and bin them now
 
      myApertureRadius = CAL_omGetApertureRadius() / real(myImageWindow%bx)

      photometricParameters%calApertureRadius =  myApertureRadius
     
      myBgrRadiusIn    = myApertureRadius + 2.0
      myBgrRadiusOut   = myApertureRadius + 6.0

  
      !************************************* 
      ! Get the Plate scale (arcsecs/pixel)
      !*************************************
      call CAL_omGetPlateScale(0.55_double, myImageWindow%plateScale)
    
!      write(*, *) "Plate scale =", myImageWindow%plateScale 

      ! Get the observation details for plotting
      call construct(myObsDetails, set = myImageSet)
      call setProgress(reporter, 7)

      ! Get the RA and DEC of the image centre
      call setLabel(reporter,"Computing sky coordinates")
      call getCelestialCentre(myCelestial, myImageWindow%RA, myImageWindow%Dec)
      call setProgress(reporter, 8)

      ! Get the pixels
      call createImageArray(myImageSet, myImage)
      call setProgress(reporter, 9)

      ! Get the quality of the pixels
      call createQualityArray(myImageSet, myQuality)
      call setProgress(reporter, 10)
      call destruct(reporter)
      ! --- finished general information


      ! Load the old sourcelist file if one has been given on the
      ! command line

 
!      oldSourceListFile = stringParameter("oldsrclistset");
  
  
      call loadSources(trim(oldSourceListFile), sources)


      mySourcesAreLoaded = associated(sources)

   end subroutine omsource_init

!*******************************************************************************
! Info: Loads the sources from the source-list file produced by omdetect & ommag
!       The 'sources' argument is populated with the detected sources.
!*******************************************************************************
subroutine omsource_loadsources(sources, sourceListFile)
    type(OmSourceParameter),  pointer    :: sources(:)
    
    character(len=256), intent(in) :: sourceListFile;
 
  
  


    call loadSources(sourceListFile, sources)
     
     



      ! Compute the RA and Declination of each source

!      call setLabel(reporter,"Celestial properties")      
    call setCelestial(myCelestial, sources)

    

end subroutine omsource_loadsources


!*************************************************************
! omsource_recompute
!
!
! Info: recomputes the properties for a source=sources(id)
!***************************************************************
subroutine omsource_recompute(sources, id)
    type(OmSourceParameter),  pointer    :: sources(:)
    integer,                  intent(in) :: id

    integer(int32) :: nx, ny
    type(OmSourceParameter),pointer :: s
    type(OmSourceParameter)         :: tmp(1)
    integer(int32)                  :: bin(2)
    character(*) , parameter :: routine="omsource_old.omsource_recompute"

    s=>sources(id)

    nx = size(myImage, 1)
    ny = size(myImage, 2)
    bin=(/myImageWindow%bx, myImageWindow%by/)

    ! Do aperture photometry on  source
    tmp(1) = s
    tmp(1)%extensionFlag = 0 ! TODO: why here, and why not before?
    call AperturePhotometry(myImage,myQuality,tmp,myApertureRadius,.true.)  
    s = tmp(1)

!write(*, *) "omsource old exposure=",  s%exposure
    ! recompute the magnitude:
    s%rawCountRate           = s%rawCountRate           / s%exposure
    s%rawCountRateError      = s%rawCountRateError      / s%exposure
    s%rawBackgroundCountRate = s%rawBackgroundCountRate / s%exposure
    call setMagnitudeProperties(s, myDeadFraction, myFrameTime, myApertureRadius)

!    s%photometryRadius = s%photometryRadius * real(myImageWindow%bx)
 
end subroutine omsource_recompute


!*************************************************************
! omsource_recompute
!
!
! Info: deletes source=sources(id) from the source file; then all sources are
!       renumbered to make the IDs consitent.
!************************************************************/
subroutine omsource_delete(sources,id)
    type(OmSourceParameter),  pointer    :: sources(:)
    integer,                  intent(in) :: id

    type(OmSourceParameter),  pointer :: v(:)
    integer                           :: l,h  ! lbound/ubound
    integer                           :: idx  ! iterator
    integer                           :: off  ! 

      ! assume that sources(id) exists
    l = maxval(lbound(sources))
    h = maxval(ubound(sources))
    off=0

    call new(v,low=l,high=h-1)
    do idx = l, h
        if (idx .eq. id) then
            off = -1
            cycle
        end if
        v(idx + off) = sources(idx)
    end do

    call delete(sources)
    sources => v
    nullify(v)

end subroutine omsource_delete


!**************************************************************
! Info: add regions to the source list
! Note: this is never called when no regions were available!
!
!**************************************************************
subroutine omsource_add(sources,regions)
    type(OmSourceParameter),  pointer    :: sources(:)
    type(OmSourceRegion),     intent(in) :: regions(:)

    ! locals
    integer                         :: i
    integer,                pointer :: idx(:)
    type(OmSourceParameter),pointer :: tmp(:)
    type(OmSourceParameter),pointer :: added(:)
    type(OmProgressT)               :: reporter    ! ... of progress to GUI

    call construct(reporter, label = "Compute added sources...", steps = 4)
    call setProgress(reporter,1)

      ! --- check new regions against current sources
    nullify(added)
    if (associated(sources)) then
        idx => getNewRegions(regions,sources,real(16/myImageWindow%bx))
        if (associated(idx)) then
            call new(added,low=1,high=size(idx))
            do i = 1, size(idx)
                added(i)%xCoord=   regions(idx(i))%x
                added(i)%yCoord=   regions(idx(i))%y
                added(i)%exposure= myObsDetails%exposure
            end do
            deallocate(idx)
        end if
        nullify(idx)
    else
        call new(added,low=1,high=size(regions))
        do i = 1, size(regions)
            added(i)%xCoord=   regions(i)%x
            added(i)%yCoord=   regions(i)%y
            added(i)%exposure= myObsDetails%exposure
!            write(*, *) "omsource_old.f90 line 382 added point " ,  added(i)%xCoord, " ",  added(i)%yCoord
        end do
    end if
    call setProgress(reporter,2)

    ! compute properties for new sources
    if (associated(added)) then
        myTrimage = TransformMatrix(0.,1.,0.,1.)
        call ComputeParameters(myImage, added, myTrimage)
        
    end if
    call setProgress(reporter, 3)

    ! merge
    if (associated(sources) .and. associated(added)) then
        
        tmp=>sources

        call new(sources, low = 1, high = size(tmp) + size(added))
        sources(1 : size(tmp)) = tmp
        do i = 1, size(added)
            sources(size(tmp) + i) = added(i)
        end do
        call delete(tmp)
        call delete(added)
    else
        ! either added or sources did not exists, so:
        if (associated(added)) sources=>added
        nullify(added)
    end if

    call setProgress(reporter,4)
    call destruct(reporter)
end subroutine omsource_add


!******************************************************
! Info: saves sources, and frees memory
! used globals:
!   myCelestial
!   myImage
!   myImageSet
!   myImageWindow
!   myObsDetails
!   myQuality
!******************************************************
subroutine omsource_finish(sources, newSourceListFile)
    type(OmSourceParameter),  pointer    :: sources(:)
    character(len=*), intent(in) :: newSourceListFile

    ! Add the sources to a table in the new source-list file
    call omsource_save(sources,&
                       newSourceListFile, &
!                         stringParameter("newsrclistset"),&
                       myImageSet,&
                       myImageWindow,&
                       myObsDetails)
    ! Close the image set, as we don't need it anymore.
    call release(myImageSet)

    ! Free memory for this module globals
    call destruct(myCelestial)
    call destruct(myImageWindow)
    call destruct(myObsDetails)
    if (associated(myImage))   deallocate(myImage)
    if (associated(myQuality)) deallocate(myQuality)

    ! And the sources themselves
    call delete(sources)
end subroutine omsource_finish



!*******************************************************************************
!--------------- helper functions private to this module -------------------
!
!
! Info: Compares the regions against the existing sources, and returns a 
!       vector containing indices to unique regions
! Note: * looks like the old MatchPoints subroutine, but cleaned-up
!       * returns a pointer, that must be destroyed by the caller
!********************************************************************************
function getNewRegions(r,s,toleranceIn) result(idx)
    type(OmSourceRegion),   intent(in)    :: r(:)
    type(OmSourceParameter),intent(in)    :: s(:)
    integer,                pointer       :: idx(:)
    real(single), optional, intent(in)    :: toleranceIn

    ! locals
    real(single)                          :: radius, tolerance
    integer                               :: i, j, nToKeep
    logical                               :: keep(size(r))
    integer                               :: status

    ! init
    nullify(idx) ! always be on the safe side ;-)
    if (present(toleranceIn)) then
        tolerance = toleranceIn
    else
        tolerance = 5.0
    end if
    nToKeep = size(keep)

    ! populate the 'keep' and count the unique points
    do i = 1, size(r)
        keep(i) = .true.
        do j = 1, size(s)
            radius = sqrt((s(j)%xCoord - r(i)%x) ** 2 +  (s(j)%yCoord - r(i)%y)** 2)
            if (radius .le. tolerance) then ! overlap
               keep(i) = .false.
               nToKeep = nToKeep - 1
               exit ! inner loop
            end if
        end do
    end do
    if (nToKeep .eq. 0) return

    ! now we create an index of all regions that are new
    allocate(idx(nToKeep), stat = status)
    call assertOn(status/=0,"omsource_old.getNewRegions.idx")

    j = 0
    do i = 1, size(r)
        if (.not.keep(i)) cycle
        j = j + 1
        idx(j) = i
    end do
end function getNewRegions

!********************************************************************
! Attempts to read a string attribute from a fits file
! (Added by CB on 14.09.2000) because apparently some attributes may
! not be present.
!********************************************************************
   subroutine GetStringAttribute(dataSet, attribute, string)
      type(DataSetT), intent(in) :: dataSet
      character(*), intent(in) :: attribute
      character(*), intent(inout) :: string

      if (.not. hasAttribute(dataSet, attribute)) then
         string = ""
      else
         string = stringAttribute(dataSet, attribute)
      end if
   end subroutine GetStringAttribute


!*********************************************************************
! Attempts to read a real attribute from a fits file
! (Added by CB on 14.09.2000) because apparently some attributes may
! not be present.
!*********************************************************************
   Function GetReal32Attribute(dataSet, attribute, defaultValue) result(number)
      type(DataSetT), intent(in) :: dataSet
      character(*), intent(in) :: attribute
      real(single), intent(in), optional :: defaultValue
      real(single) :: number

      if (.not. hasAttribute(dataSet, attribute)) then
         if (present(defaultValue)) then
            number= defaultValue
         else
            number = 0.0
         end if
      else
         number = real32Attribute(dataSet, attribute)
      end if

   end Function GetReal32Attribute


!**************************************************
! Gets the 2-d image from the fits file
!**************************************************
subroutine createImageArray(dataSet,image)
    type(dataSetT), intent(in) :: dataSet
    real(single),   pointer    :: image(:, :)

    type(ArrayT)           :: handle
    integer(int8), pointer :: int8ImageP(:, :)     ! Integer8 array pointer.
    integer(int32),pointer :: int32ImageP(:, :)    ! Integer32 array pointer.
    integer(int16),pointer :: int16ImageP(:, :)    ! Integer16 array pointer.
    real(single),  pointer :: real32ImageP(:, :)   ! Real32 array pointer.
    real(double),  pointer :: real64ImageP(:, :)   ! Real64 array pointer.
    integer                :: sz(2)
    integer                :: status

    nullify(image)
    handle = array(dataSet, "PRIMARY", read)

    sz = dimensions(handle)
    if (all(sz > 0)) then
        allocate( image(0:sz(1)-1,0:sz(2)-1),stat=status)
        call assertOn(status/=0,"omsource_old.createImageArray.image")

        select case(arrayDataType(handle))
        case(integer8)
            int8ImageP => int8Array2Data(handle)
            image = real(int8ImageP)

        case(integer16)
            int16ImageP => int16Array2Data(handle)
            image = real(int16ImageP)

        case(integer32)
            int32ImageP => int32Array2Data(handle)
            image = real(int32ImageP)

        case(real32)
            real32ImageP => real32Array2Data(handle)
            image = real32ImageP

        case(real64)
            real64ImageP => real64Array2Data(handle)
            image = real64ImageP
        end select
    end if

    call release(handle)
end subroutine createImageArray
!*********************************************************
!
!*********************************************************


subroutine createQualityArray(dataSet,quality)
    type(dataSetT), intent(in) :: dataSet
    logical(bool),  pointer    :: quality(:, :)

    type(ArrayT)                :: handle
    integer(int8), pointer :: arrayP(:, :)
    integer(int32)              :: sz(2)
    integer                     :: status

    nullify(quality)
    handle = array(dataSet, "QUALITY", read)

    sz = dimensions(handle)
    if (all(sz > 0)) then
        arrayP => int8Array2Data(handle)

        allocate( quality(0:sz(1)-1, 0:sz(2)-1), stat=status)
        call assertOn(status/=0,"omsource_old.createQualityArray.quality")

        where (arrayP .ne. 0)
            quality = .true.
        elsewhere
            quality = .false.
        end where
    end if

    call release(handle)
end subroutine createQualityArray


!*********************************************************************************
! subroutine loadSources
! Info: loads sources found in dataset(setName), and returns a newly created vector
!       of sources.
! 
!********************************************************************************
subroutine loadSources(setName, sources)
    character(*),            intent(in) :: setName
    type(OmSourceParameter), pointer    :: sources(:)

    ! locals
    integer(int32)        :: nSources
    real(single)          :: exposure 
    type(DataSetT)        ::  set
    type(TableT)          :: inTableH
          

    type(OmProgressT)     :: reporter    ! ... of progress to GUI    

      ! --- don't bother if none available
    if (trim(setName) == "" ) return

    ! Read in the OM OSW Source List file.
    set = dataSet(setName, read)
    exposure = GetReal32Attribute(set, "EXPOSURE", 0.0)  
 
    if (.not. hasBlock(set, "SRCLIST")) then
        call destruct(reporter)

        ! Incompatible file - can't continue
        write (errstr, *) "Could not find any compatibile blocks in set ", trim(setName)
        call fatal("BadFileType")

    else
        ! Get a handle on the source list binary table extension.
        inTableH = table(set, "SRCLIST")

    end if

     
    exposure = GetReal32Attribute(set, "EXPOSURE", 0.0) 


    ! Get the number of rows in the source list binary table extension.
    nSources = numberOfRows(inTableH)

    if (nSources .le. 0) then
        call destruct(reporter)
        write (errstr, *) "No sources found in dataset: ", trim(setName)

        call fatal("NoSources")
    end if



    ! Get the limiting magnitude of the image
    myDetectionRate      = GetReal32Attribute(set, "CRLIMIT")
    myDetectionMagnitude = GetReal32Attribute(set, "MAGLIMIT")
     


    call construct(reporter, label="Loading sourcelist data", steps = 23)

    call delete(sources)
    call new(sources, low = 1, high = nSources)
 

    !*********************************************************
    ! Get the data for each source from the source list file
    !*********************************************************
    call setProgress(reporter, 1)
  
    call LoadReal32Column(inTableH, "XPOS",  sources%xCoord)
    call setProgress(reporter, 2)
  
    call LoadReal32Column(inTableH, "YPOS",  sources%yCoord)
    call setProgress(reporter, 3)
  
    call LoadReal32Column(inTableH, "FWHM_MAJ",  sources%semiA)
    call setProgress(reporter, 4)
  
    call LoadReal32Column(inTableH, "FWHM_MIN",  sources%semiB)
    call setProgress(reporter, 5)
  
    call LoadReal32Column(inTableH, "PA", sources%paSource)
    call setProgress(reporter, 6)
  

    call LoadReal32Column(inTableH, "RATE", sources%rawCountRate)
    call setProgress(reporter, 7)
  
    call LoadReal32Column(inTableH, "RATE_ERR", sources%rawCountRateError)
    call setProgress(reporter,8)

    call LoadReal32Column(inTableH, "CORR_RATE", sources%countRate)
    call setProgress(reporter, 9)
  
    call LoadReal32Column(inTableH, "CORR_RATE_ERR", sources%countRateError)
    call setProgress(reporter, 10)
  
    call LoadReal32Column(inTableH, "COINCLOSS_CORR", sources%coincidenceCorrection)
    call setProgress(reporter,11)
    call LoadReal32Column(inTableH, "PSF1_CORR", sources%psf1Correction)
    call setProgress(reporter, 12)
  
    call LoadReal32Column(inTableH, "PSF2_CORR", sources%psf2Correction)
    call setProgress(reporter,13)

    call LoadReal32Column(inTableH, "BACKGROUND_RATE",  sources%rawBackgroundCountRate)
    call setProgress(reporter, 14)
  
    call LoadReal32Column(inTableH, "CORR_BACKGROUND_RATE",  sources%backgroundCountRate)
    call setProgress(reporter,15)

    call LoadReal32Column(inTableH, "MAG", sources%magnitude)
    call setProgress(reporter, 16)
  
    call LoadReal32Column(inTableH, "MAGERR", sources%magnitudeError)
    call setProgress(reporter, 17)
  
    call LoadReal32Column(inTableH, "SIGNIFICANCE", sources%significance)
    call setProgress(reporter, 18)
  
    call LoadReal32Column(inTableH, "APERTURE", sources%photometryRadius)

    !*****************************************************************
    ! Need binned aperture radius
    !*****************************************************************

    sources%photometryRadius = sources%photometryRadius / myImageWindow%bx
    call setProgress(reporter, 19)

    ! Get the RAS and decs if they exist
    call LoadReal64Column(inTableH, "RA", sources%rightAscension)
    call setProgress(reporter, 20)
  
    call LoadReal64Column(inTableH, "DEC", sources%declination)
    call setProgress(reporter, 21)
  
    call LoadInt8Column(inTableH, "EFLAG", sources%extensionFlag)


    call setProgress(reporter,22)

    call release(inTableH)
    call release(set)

    sources%innerBackgroundRadius = myBgrRadiusIn

    sources%outerBackgroundRadius = myBgrRadiusOut
    sources%exposure = exposure 
    call setProgress(reporter,23)

    call destruct(reporter)

end subroutine loadSources

!****************************************************************************
! Returns an array containing the data in the named column of the given table
!
!****************************************************************************
subroutine  LoadReal32Column(inTableH, name, data)
      type(TableT), intent(in) :: inTableH
      character(*), intent(in) :: name
      real(single), intent(out) :: data(:)

      type(ColumnT)         :: columnH
      real(single), pointer :: dataPtr(:) ! Magnitude error pointer.
      integer(int32)        :: i, sz
     

     
      if (hasColumn(inTableH, name)) then
         columnH = column(inTableH, name)
         dataPtr => real32Data(columnH)
         sz = numberOfRows(inTableH)

      


         do i = 1, sz
            data(i) = dataPtr(i - 1)
           
         end do
         call release(columnH)
     else
        call warning("No such data column ", name)


      end if
   end subroutine LoadReal32Column

!******************************************************************************
! Returns an array containing the data in the named column of the given table
!******************************************************************************
   subroutine LoadReal64Column(inTableH, name, data)
      type(TableT), intent(in) :: inTableH
      character(*), intent(in) :: name
      real(double), intent(out) :: data(:)

      type(ColumnT)         :: columnH
      real(double), pointer :: dataPtr(:) ! Magnitude error pointer.
      integer(int32)        :: i, sz
      

     
      if (hasColumn(inTableH, name)) then
         columnH = column(inTableH, name)
         dataPtr => real64Data(columnH)
         sz = numberOfRows(inTableH)

         
       

         do i = 1, sz
            data(i) = dataPtr(i - 1)
         end do
         call release(columnH)
      end if
   end subroutine LoadReal64Column


!***************************************************************************
!
! Returns an array containing the data in the named column of the given table
!***************************************************************************
   subroutine LoadInt8Column(inTableH, name, data) 
      type(TableT), intent(in) :: inTableH
      character(*), intent(in) :: name
      integer(int8), intent(out) :: data(:)

      type(ColumnT)         :: columnH
      integer(int8), pointer :: dataPtr(:) ! Magnitude error pointer.
      integer(int32)        :: i, sz
     

   
      if (hasColumn(inTableH, name)) then
         columnH = column(inTableH, name)
         dataPtr => int8Data(columnH)
         sz = numberOfRows(inTableH)

       

         do i = 1, sz
            data(i) = dataPtr(i - 1)
         end do
         call release(columnH)
      end if
   end subroutine LoadInt8Column


!**********************************************************
! Info: saves the sources into dataset(fileName)
! Used globals:
!   mySourcesAreLoaded
!   myDetectionRate
!   myDetectionMagnitude
!   myFrameTime
!   myDeadFraction
!   myNullValue
! TODO: save unbinnend radius? save baground radii?
!*********************************************************
subroutine omsource_save(sources, fileName, imageSet, iw, od)
      type(OmSourceParameter), pointer    :: sources(:)
      character(*),            intent(in) :: fileName
      type(DataSetT),          intent(in) :: imageSet
      type(ImageWindow),       intent(in) :: iw
      type(ObsDEtails),        intent(in) :: od

      type(DataSetT)         :: set
      type(tableT)           :: t
      real(double)           :: plateScale
      integer                :: i, status, nSources
      real(single),  pointer :: temp(:)
      real(double),  pointer :: gLon(:), gLat(:)
      integer(int32),pointer :: number(:)
      type(OmProgressT)      :: reporter    ! ... of progress to GUI
      character(*) , parameter :: routine="omsource_old.omsource_save"

      if (.not.associated(sources)) return ! don't bother
      nSources = size(sources)

      sources%photometryRadius = sources%photometryRadius *  real(myImageWindow%bx) 

      ! notify the GUI that we are busy...
      call construct(reporter, label="Saving source list", steps = 20)
      call setProgress(reporter, 1)

      set = dataSet(trim(fileName), create)
      call setProgress(reporter,2)

      ! --- set all necessary attributes
      ! copy the keywords from the image header
      call copyAllAttributes(attributable(imageSet), attributable(set))
      call setProgress(reporter,3)

      ! now set some defaults attributes
      call setAttribute(set,"WINDOWX0", iw%x0,    "pixels","Window x-origin")
      call setAttribute(set,"WINDOWDX", iw%width, "pixels","Window dx")
      call setAttribute(set,"WINDOWY0", iw%y0,    "pixels","Window y-origin")
      call setAttribute(set,"WINDOWDY", iw%height,"pixels","Window dy")
      call setAttribute(set,"BINBPE",  iw%binBpe,"Window BPE set?")
      call setAttribute(set,"OBS_ID",  trim(od%obsID),"Observation identifier")
      call setAttribute(set,"OBJECT",  trim(od%object),"Object")
      call setAttribute(set,"TELESCOP", trim(od%telescope),"Telescope")
      call setAttribute(set,"OBSERVER", trim(od%observer),"Observer")
      call setAttribute(set,"DATE-OBS", trim(od%startObs),"Start time of observations")
      call setAttribute(set,"DATE-END", trim(od%endObs),"End time of observations")
      call setAttribute(set,"EXPOSURE", od%exposure,"secs","Exposure duration")
      call setAttribute(set,"RA_PNT", iw%ra,"degrees","RA pointing direction")
      call setAttribute(set,"DEC_PNT", iw%dec,"degrees","Dec pointing direction")

      ! this is an osw source list
      call setAttribute(set,"CONTENT","OM OSW SOURCE LIST",&
           "File contents")
      call setAttribute(set,"HDUCLASS","OGIP",&
           "Format conforms to OGIP/GSFC conventions")
      call setAttribute(set,"HDUCLAS1","SRCLIST",&
           "File contains a source list")
      call setProgress(reporter,4)

      ! add the detection rate and limiting magnitude if necessary
      if (.not. mySourcesAreLoaded) then
         call ComputeLimitingMagnitude(sources)
         call setAttribute(set, "CRLIMIT", myDetectionRate, "counts/sec", "Detection limit count rate")
         call setAttribute(set, "MAGLIMIT", myDetectionMagnitude, "mag", "detection limiting magnitude")
      end if
 
     
     


     
      ! Add various CCD parameters to the header
!      call setAttribute(set, "FRAMTIME", (myFrameTime * 1000.0), "millisecs", "CCD frame time")
!      call setAttribute(set, "DEADFRAC", myDeadFraction, "", "CCD dead fraction")
      call setProgress(reporter,5)
      
      ! Now add the columns one by one starting with the source number
      t = addTable(set, "SRCLIST", nSources)
      allocate(number(1:nSources),stat=status)
      call assertOn(status/=0,routine//".number")
      do i=1,nSources
         number(i) = i
      end do
      call  AddInt32Column(t, "SRCNUM", " ", nSources, number)
      deallocate(number)
      call setProgress(reporter,6)


      ! Add the source positions
      call AddReal32Column(t, "XPOS", "pixels", "Centroided x-source position ", &
           nSources, sources%xCoord, myNullValue)
      call AddReal32Column(t, "YPOS", "pixels", "Centroided y-source position", &
           nSources, sources%yCoord, myNullValue)
      
 

     call  AddReal64Column(t, "RA", "degrees", " ", &
           nSources, sources%rightAscension, myNullValue)
      call  AddReal64Column(t, "DEC", "degrees", " ", &
           nSources, sources%declination, myNullValue)
      call setProgress(reporter,7)


      ! add the galactic longitude/latitude
      allocate(gLon(1:nSources),stat=status)
      call assertOn(status/=0,routine//".gLon")
      allocate(gLat(1:nSources),stat=status)
      call assertOn(status/=0,routine//".gLat")
      do i = 1, nSources
         call sla_EQGAL(sources(i)%rightAscension, sources(i)%declination, &
              gLon(i), gLat(i))
         gLon(i) = gLon(i) * radiansToDegrees
         gLat(i)  = gLat(i)  * radiansToDegrees
      end do

      call  AddReal64Column(t, "LII", "degrees", " ", &
           nSources, gLon, myNullValue)
      call  AddReal64Column(t, "BII", "degrees", " ", &
           nSources, gLat, myNullValue)
      deallocate(gLon)
      deallocate(gLat)
      call setProgress(reporter,8)

      sources%xCoordError = 0.5 ! TODO: hmmm?
      call AddReal32Column(t, "POSERR", "pixels", " ", &
           nSources, sources%xCoordError, myNullValue)
      call setProgress(reporter,9)


      ! Add the sources-count rates
      call AddReal32Column(t, "RATE", "counts/sec", &
           "No corrections have been applied", &
           nSources, sources%rawCountRate, myNullValue)
     call AddReal32Column(t, "RATE_ERR", "counts/sec", "No corrections have been applied", &
           nSources, sources%rawCountRateError, myNullValue)

      ! Add the sources background rate
    call AddReal32Column(t, "BACKGROUND_RATE",  "counts/sec", &
            "No corrections have been applied", &
           nSources, sources%rawBackGroundCountRate, myNullValue)
  

    call AddReal32Column(t, "SIGNIFICANCE", "sigma", "(source-background)/sqrt(background)", &
           nSources, sources%significance, myNullValue)
  

     call setProgress(reporter,10)

    call setProgress(reporter,11)


      ! Add the columns for the corrected count rates,
      call AddReal32Column(t, "CORR_RATE", "counts/sec", &
           "Corrected for coincidence loss, etc", &
           nSources, sources%countRate,  myNullValue)
      call AddReal32Column(t, "CORR_RATE_ERR", "counts/sec",  &
          "Corrected for coincidence loss, etc", &
           nSources, sources%countRateError, myNullValue)

      call AddReal32Column(t, "CORR_BACKGROUND_RATE", "counts/sec", &
           "Corrected for coincidence loss, etc", &
           nSources, sources%backgroundCountRate, myNullValue)
 


      call setProgress(reporter,12)

      call AddReal32Column(t, "PSF1_CORR", "%", &
           "PSF correction from measurement to 12 pixel aperture",  &
           nSources, sources%psf1Correction, myNullValue)
      call AddReal32Column(t, "PSF2_CORR", "%", &
           "PSF correction from 12 to 35 pixels",  &
           nSources, sources%psf2Correction, myNullValue)
        call setProgress(reporter,13)
     call AddReal32Column(t, "COINCLOSS_CORR", "%", &
           "Coincidence-loss correction", &
           nSources, sources%coincidenceCorrection, myNullValue)
  
    call AddReal32Column(t, "APERTURE", "unbinned pixels", &
           "Aperture radius used in photometry ", &
           nSources,sources%photometryRadius, myNullValue)
 

      ! Add the sources instrumental magnitudes
      call AddReal32Column(t, "MAG", "magnitudes", &
           "Instrumental magnitude", &
           nSources, sources%magnitude, myNullValue)
      call AddReal32Column(t, "MAGERR", "magnitudes", &
           "Instrumental magnitude error", &
           nSources, sources%magnitudeError, myNullValue)
      call setProgress(reporter,14)

    

      ! Convert width to arcsecs
      call CAL_omGetPlateScale(0.55_double, plateScale)
      sources%semiA = sources%semiA * iw%bx * plateScale
      sources%semiB = sources%semiB * iw%by * plateScale
      call setProgress(reporter,15)

      allocate(temp(1 : nSources),stat=status)
      call assertOn(status/=0,routine//".temp")
      do i = 1,nSources
         temp(i) = 2.354 * max(sources(i)%semiA,sources(i)%semiB)
      end do
      call AddReal32Column(t, "FWHM_MAJ", "arcsecs", " ", &
           nSources, temp, myNullValue)
      call setProgress(reporter,16)

      do i = 1,nSources
         temp(i) = 2.354 * min(sources(i)%semiA,sources(i)%semiB)
      end do
      call AddReal32Column(t, "FWHM_MIN", "arcsecs", " ", &
           nSources, temp, myNullValue)
      deallocate(temp)
      call setProgress(reporter,17)

      call AddReal32Column(t, "PA", "degrees", " ", &
           nSources, sources%paSource, myNullValue)
      call setProgress(reporter,18)


      ! Add the quality flags
      call AddInt8Column(t, "QFLAG", "none", nSources, sources%qualityFlag)
      call AddInt8Column(t, "CFLAG", "none", nSources, sources%confusionFlag)
      call AddInt8Column(t, "EFLAG", "none", nSources, sources%extensionFlag)
      call setProgress(reporter,19)

      call release(t)
      call release(set)
      call setProgress(reporter,20)

      call destruct(reporter)

end subroutine omsource_save


!****************************************************************
! Info: computes all properties of the sources specified
! Note: * Each input source is 'empty' save the (xCoord,yCoord) properties
!       * 'trimage' argument is a residual, and should be removed
!****************************************************************

subroutine computeParameters(image, sources, trimage)
    real(single),           pointer    :: image(:, :)
    type(OmSourceParameter),pointer    :: sources(:)
    real(single),           intent(in) :: trimage(1 : 6)

    integer(int32)                     :: i,nSources,nx, ny
    integer(int32)                     :: i0, il, iu, j0, jl, ju, dx
    real(single) :: backgroundMode
    nx=size(image, 1)
    ny=size(image, 2)
    nSources = size(sources)
    dx = 12 / myImageWindow%bx

      backgroundMode = 0.0

      ! Compute the centre coordinates of the new sources
      do i = 1, nSources
        
   
         call computeCentreOfSource(nx,ny, image, trimage, &
              sources(i)%xCoord, sources(i)%yCoord, &
              i0, il, iu, j0, jl, ju)
         il = i0 - dx
         il = max(1, il)
         iu = i0 + dx
         iu = min(iu, nx)
         jl = j0 - dx
         jl = max(1, jl)
         ju = j0 + dx
         ju = min(ju, ny)

         call setImageCoordProperties(sources(i), image,  backgroundMode, il, iu, jl, ju)
      end do

      call photometry(image, nSources, sources) ! see below

      ! Compute the RA and Declination of each source
      call setCelestial(myCelestial, sources)
end subroutine computeParameters

!************************************************************
! Subroutine photometry
!***********************************************************
subroutine photometry(image, nSources,  sources)
    real(single),           pointer    :: image(:, :)
    integer(int32),         intent(in) :: nSources
    type(OmSourceParameter),pointer    :: sources(:)

    integer(int32)         :: i
    integer(int32)         :: nx, ny
    integer(int32)         :: bin(2)
    type(OmSourceParameter),pointer :: s

    nx = size(image, 1)
    ny = size(image, 2)

      ! initialise defaults:
    do i = 1, nSources
        s=>sources(i)
        s%photometryRadius      = myApertureRadius
        s%innerBackgroundRadius = myBgrRadiusIn
        s%outerBackgroundRadius = myBgrRadiusOut
        s%exposure              = myObsDetails%exposure
        s%extensionFlag         = 1
    end do

    ! Do aperture photometry on each sources
    bin = (/myImageWindow%bx, myImageWindow%by/)

   


    call AperturePhotometry(myImage, myQuality, sources, myApertureRadius)


    ! Convert counts to count rates, and set the magnitude
    do i = 1, nSources
        s=>sources(i)

        s%rawCountRate           =  s%rawCountRate / s%exposure
        s%rawCountRateError      =  s%rawCountRateError / s%exposure
        s%rawBackGroundCountRate =  s%rawBackGroundCountRate / s%exposure


        call setMagnitudeProperties(s, myDeadFraction, myFrameTime, myApertureRadius)
    end do
end subroutine photometry

!********************************************
! AddReal32Column
!********************************************

   subroutine AddReal32Column(tableSet, heading,  &
        units, comment, sz, array, myNullValue, position)
      type(TableT),   intent(in) :: tableSet
      character(*),   intent(in) :: heading, units, comment
      integer,        intent(in) :: sz
      real(single),   intent(in) :: array(0:sz-1)
      real(single),   intent(in) :: myNullValue
      integer(int32), intent(in), optional :: position

      real(single), pointer :: real32Col(:)
      type(ColumnT)         :: pColumn
      integer               :: i
      integer               :: ndims(1) = (/1/)


      if (.not. hasColumn(tableSet, trim(heading))) then
         if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), real32, &
                 trim(units), ndims, trim(comment), position)
         else
            pColumn = addColumn(tableSet, trim(heading), real32, &
                 trim(units), ndims, trim(comment))
         end if
      else
         pColumn = column(tableSet, trim(heading), modify)
      end if


      real32Col => real32Data(pcolumn)
      do i = 0 , sz - 1
         if (array(i) .eq. myNullValue) then
            call setToNull(pcolumn, i)
         else
            real32Col(i) = array(i)
         end if

      enddo
      call release(pcolumn)

   end subroutine AddReal32Column


   ! AddReal32Cololumn




   subroutine AddReal64Column(tableSet, heading,  &
        units, comment, sz, array, myNullValue, position)

      type(TableT), intent(in) :: tableSet
      character(*), intent(in) :: heading, units, comment
      integer, intent(in) :: sz
      real(double), intent(in), dimension(0 : sz - 1) :: array
      real(single), intent(in) :: myNullValue
      integer(int32), intent(in), optional :: position
      real(double), pointer :: real64Col(:)
      type(ColumnT) :: pColumn
      integer i
      integer :: ndims(1) = (/1/)


      if (.not. hasColumn(tableSet, trim(heading))) then
         if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), real64, trim(units), ndims, trim(comment), position)
         else
            pColumn = addColumn(tableSet, trim(heading), real64, trim(units), ndims, trim(comment))
         end if
      else
         pColumn = column(tableSet, trim(heading), modify)
      end if


      real64Col => real64Data(pcolumn)

      do i = 0 , sz - 1
         if (array(i) .eq. dble(myNullValue)) then
            call setToNull(pcolumn, i)
         else
            real64Col(i) = array(i)
         end if

      enddo
      call release(pcolumn)

   end subroutine AddReal64Column



   ! AddInt8Column
   ! Adds a column of integer8 data to a fits table.

   subroutine AddInt8Column(tableSet, heading, comment, &
        sz, array)

      type(TableT), intent(in) :: tableSet
      character(*), intent(in) :: heading, comment
      integer(int32), intent(in) :: sz
      integer(int8), intent(in), dimension(:) :: array
      integer(int8), pointer :: int8Col(:)
      type(ColumnT) :: pColumn
      integer(int32) :: i
      integer(int32):: ndims(1) = (/1/)

      pcolumn = addColumn(tableSet, heading, integer8, comment, ndims)
      int8Col => int8Data(pcolumn)

      do i =1,sz
         int8Col(i - 1) = array(i)
      end do
      call release(pcolumn)
   end subroutine AddInt8Column


   ! AddInt8Column
   ! Adds a column of integer8 data to a fits table.

   subroutine AddInt32Column(tableSet, heading, comment, &
        sz, array)
      type(TableT), intent(in) :: tableSet
      character(*), intent(in) :: heading, comment
      integer(int32), intent(in) :: sz
      integer(int32), intent(in), dimension(:) :: array
      integer(int32), pointer :: int32Col(:)
      type(ColumnT) :: pColumn
      integer(int32) :: i
      integer(int32):: ndims(1) = (/1/)

      pcolumn = addColumn(tableSet, heading, integer32, comment, ndims)
      int32Col => int32Data(pcolumn)

      do i =1,sz
         int32Col(i - 1) = array(i)
      end do
      call release(pcolumn)

   end subroutine AddInt32Column



   ! Info: find a more accurate centre pixel for (xPos,yPos)
   subroutine computeCentreOfSource(nx, ny, image, trimage, xPos, yPos, i0, il, iu, j0, jl, ju)
      integer(int32), intent(in) :: nx, ny
      real(single), intent(in) :: image(1 : nx, 1 : ny)
      real(single), intent(in) :: trimage(1 : 6)
      real(single), intent(in) :: xPos, yPos

      integer(int32), intent(out) :: i0, il, iu, j0, jl, ju
      integer(int32) :: i, j
      real(single) :: maximum

      ! current pixel centre
      i0 = Ival(xPos, trimage)
      j0 = Jval(yPos, trimage)


      ! now find a more accurate centre for the image
      il = i0 - 5
      il = max (il, 1)
      iu = il + 10
      iu = min (iu, nx)

      jl = j0 - 5
      jl = max (jl, 1)
      ju = jl + 10
      ju = min (ju, ny)
      maximum = 1e-6
      do i = il, iu
         do j = jl, ju
            if (image(i, j) .gt. maximum) then
               maximum = image(i, j)
               i0 = i
               j0 = j
            end if
         end do
      end do


      il = i0 - 25
      il = max (il, 1)
      iu = il + 50
      iu = min (iu, nx)

      jl = j0 - 25
      jl = max (jl, 1)
      ju = jl + 50
      ju = min (ju, ny)
   end subroutine computeCentreOfSource



   subroutine construct_od(o,set)
      type(ObsDetails),  intent(out)    :: o
      type(DataSetT),    intent(in)     :: set
      
      call GetStringAttribute(set, "TELESCOP", o%telescope)
      call GetStringAttribute(set, "OBSERVER", o%observer)
      call GetStringAttribute(set, "OBJECT",   o%object)
      call GetStringAttribute(set, "FILTER",   o%filter)
      call GetStringAttribute(set, "INSTRUME", o%instrument)
      call GetStringAttribute(set, "OBS_ID",   o%obsID)
      call GetStringAttribute(set, "EXP_ID",   o%expID)
      call GetStringAttribute(set, "DATE-OBS", o%startObs)
      call GetStringAttribute(set, "DATE-END", o%endObs)
      o%exposure=GetReal32Attribute(set, "EXPOSURE")
   end subroutine construct_od



   subroutine destruct_od(o)
      type(ObsDetails), intent(inout)    :: o
      o%telescope=''
      o%exposure=0.
   end subroutine destruct_od



   subroutine construct_iw(w,set)
      type(ImageWindow), intent(out)    :: w
      type(DataSetT),    intent(in)     :: set
      
      type(TableT)                      :: t
      type(ColumnT)                     :: c
      integer(int32), pointer            :: ptr(:) ! release on table release
      


      t = table(set, "MODES")
      c = column(t, "WINDOWX0", read)
      ptr => int32Data(c)
      w%x0  = ptr(0)
      c = column(t, "WINDOWY0", read)
      ptr => int32Data(c)
      w%y0  = ptr(0)
      c = column(t, "WINDOWDX", read)
      ptr => int32Data(c)
      w%width  = ptr(0)
      c = column(t, "WINDOWDY", read)
      ptr => int32Data(c)
      w%height  = ptr(0)
      call release(t)

      ! compute binning
      w%binBpe = booleanAttribute(set, "BINBPE")
      w%bx     = 2**int32Attribute(set, "BINAX1")
      w%by     = 2**int32Attribute(set, "BINAX2")
      if (w%binBpe) then
         w%bx=w%bx*2
         w%by=w%by*2
      end if

      w%filter  = stringAttribute(set, "FILTER")
      w%datamode= stringAttribute(set,"DATAMODE")


      !*********************************************************
      ! CCD frame-time and dead-fraction
      !*********************************************************
     
      w%frameTime = real64Attribute(set, "FRAMTIME") / 1000.0
      w%deadFraction = real64Attribute(set, "DEADFRAC")
    

   end subroutine construct_iw

   subroutine destruct_iw(w)
      type(ImageWindow), intent(inout)    :: w
      w%filter=''
      w%datamode=''
   end subroutine destruct_iw


   ! -globals modified:
   !  mySourceMap (allocated/deallocated here!)
   !  myDetectionRate
   !  myDetectionMagnitude
   !
   ! -globals used:
   !  myImage
   !  myApertureRadius
   !  mySourceMap
   !  myObsDetails
   !  myImageWindow
   !  myDeadFraction, myFrameTime
   !  myApertureRadius
   subroutine ComputeLimitingMagnitude(sources)
      type(OmSourceParameter),  pointer :: sources(:)

      type(OmSourceParameter) :: source
      real(single)            :: sigma
      integer(int16), pointer :: map(:,:)
      if (.not.associated(sources)) return

      ! init
      myDetectionRate      = 0.0
      myDetectionMagnitude = 0.0

      ! Compute the background mode of the image      
      map=>createIdMap(myImage,sources,myApertureRadius)  
      myBgrMode = computeBackgroundMode(myImage,map,3.0,sigma)
      deallocate(map)
      nullify(map)

      if (myBgrMode.gt. 0.0) then  ! was boxScale**2 -> now the aperture area
         myDetectionRate = myApertureRadius*&
              sqrt(myBgrMode*PI*myApertureRadius**2)/myObsDetails%exposure

         ! Compute the detection limiting magnitude of the centre of the image
         call construct(source)
         source%xCoord                = real(size(myImage, 1)) / 2.0
         source%yCoord                = real(size(myImage, 2)) / 2.0
         source%photometryRadius      = myApertureRadius
         source%rawCountRate          = myDetectionRate
         source%rawCountRateError     = sqrt(myDetectionRate)
         source%rawBackgroundCountRate= 0.
         source%exposure              = myObsDetails%exposure
         call setMagnitudeProperties(source, myDeadFraction, myFrameTime, myApertureRadius)
         myDetectionMagnitude = source%magnitude
      end if

      write(errstr,*) "Detection Magnitude : ", myDetectionMagnitude
      call message(AppMsg, Verbose)
      write(errstr,'(a, 1pg9.3)') "Detection Count Rate: ", myDetectionRate
      call message(AppMsg, Verbose)

   end subroutine ComputeLimitingMagnitude



   !--------------------------------------------------------------------------!
   !                    other helpers                                         !
   !--------------------------------------------------------------------------!

   ! Sets up plotting transformation matrix.
   ! Added by CB to prevent the code having to be 
   ! repeated twice in the main program 
   ! (as in original).
   function TransformMatrix(xmin, xbinWidth, &
        ymin, ybinWidth) result(trimage)
      real(single), intent(in) :: xmin
      real(single), intent(in) :: xbinWidth
      real(single), intent(in) :: ymin
      real(single), intent(in) :: ybinWidth
      real(single) :: trimage(6)

      trimage(1) = xmin - xbinWidth / 2.0
      trimage(2) = xbinWidth
      trimage(3) = 0.0
      trimage(4) = ymin - ybinWidth / 2.0
      trimage(5) = 0.0
      trimage(6) = ybinWidth

   end function TransformMatrix

   ! Returns i for a given x
   integer(int32) function Ival(x, trimage)
      real(single), intent(in) :: x
      real(single), intent(in) :: trimage(1 : 6)

      Ival = 1 + int((x - trimage(1)) / trimage(2))
   end function Ival


   ! Returns j for a given y
   integer(int32) function Jval(y, trimage)
      real(single), intent(in) :: y
      real(single), intent(in) :: trimage(1 : 6)

      Jval = 1 + int((y - trimage(4)) / trimage(6))   
   end function Jval

end module omsource_old
