module omsourcebackground_mod
   use cal
   use dal     

   use omsourceutils_mod
   implicit none

   private  ! all,but:
   public :: computeBackgroundMode
   public :: computeBackgroundMode1

   character(*), parameter :: myModule="omsourcebackground_mod"

contains

   ! Info: Computes the background mode (counts/per pixel) using the image
   !       (background + source) and the source-map image (levelImage). Ignores
   !       any image pixels that correspond to a levelImage one that has a
   !       non-zero value (ie a source pixel) Also returns the variance
   !       of the sky-background 
   function computeBackgroundMode(image, level, nSigma, sigma) result(mode)
      real(single),   pointer     :: image(:,:)
      integer(int16), pointer     :: level(:,:)
      real(single),   intent(in)  :: nSigma
      real(single),   intent(out) :: sigma
      real(single)                :: mode

      integer(int32)              :: i, j
      integer(int32)              :: nlist
      real(single),   pointer     :: list(:)
      integer                     :: status
      integer(int32)              :: lb(2),ub(2)
      character(*),   parameter   :: routine=myModule//".computeBackgroundMode"

      lb=lbound(image)
      ub=ubound(image)

      allocate(list(size(image)),stat=status)
      call assertOn(status/=0,routine//".list") 

      nlist = 0
      do i = lb(1),ub(1)
         do j = lb(2),ub(2)

            if (level(i, j) .eq. 0) then
               nList = nList + 1
               list(nlist) = image(i, j) 
            end if

         end do
      end do

      ! only pass the valid subsection
      mode = backgroundMode(list(1:nlist), 0.2, nSigma, sigma)
      deallocate(list)
   end function computeBackgroundMode



   ! Info: Computes the background mode counts using the image
   !       (background + source) 
   function computeBackgroundMode1(background,nSigma,stddev) result(mode)
      real(single),   intent(in)  :: background(:)
      real(single),   intent(in)  :: nSigma
      real(single),   intent(out) :: stddev
      real(single)                :: mode

      mode = backgroundMode(background, 0.2, nSigma, stddev)
   end function computeBackgroundMode1



   !--------------- helper functions private to this module -------------------



   ! Info: backgroundMode() ??? TODO ...
   function backgroundMode(data, tolerance, nSigma, sigma) result(mode)
      real(single),   intent(in) :: data(:)
      real(single),   intent(in) :: tolerance ! Tolerance is usually set to 0.2
      real(single),   intent(in) :: nSigma
      real(single),   intent(out):: sigma

      real(single),   pointer    :: list(:)
      integer(int32)             :: nlist
      real(double)               :: mean
      real(single)               :: mode
      integer(int32)             :: i, k
      real(single)               :: median
      real(double)               :: variance, originalSigma, testSigma
      real(single)               :: lowValue, highValue
      integer(int32)             :: status, m
      character(*), parameter :: routine=myModule//".backgroundMode"


      ! create the sorted list of data points:
      nList = size(data)
      allocate(list(nList),stat=status)
      call assertOn(status/=0,routine//".list")
      list = sorted(data)
      mode = medianOfSorted(list)

      ! Get variance and std deviation
      variance = 0.0_double
      do k = 1,nList
         variance = variance + dble((list(k) - mode) ** 2)
      end do

      variance      = variance / real(nList - 1)
      originalSigma = sqrt(variance)
      lowValue      = mode - nSigma * originalSigma
      highValue     = mode + nSigma * originalSigma


      ! reject data at more than nSigma from data
      testSigma = originalSigma
      sigmaClip: do i = 1, 50
         m = 0
         mean = 0.0_double
         do k = 1, nlist
            if(list(k) .ge. lowValue .and. list(k) .le. highValue) then
               m = m + 1
               list(m) = list(k)
               mean = mean + dble(list(m))
            end if
         end do
         nlist = m
         if(nlist .lt. 20) exit sigmaClip

         mean   = mean / dble(nList)     
         median = medianOfSorted(list(1:nList))
         mode   = 2.5 * median - 1.5 * mean
         if (mode .le. 0.0 .or. mode .gt. 2.0 * median) mode = median


         ! Now do sigma clipping; get variance and std deviation
         variance = 0.0_double
         do k = 1,nList
            variance = variance + dble((list(k) - mode) ** 2)
         end do

         variance  = variance / (nlist - 1)
         sigma     = sqrt(variance)
         lowValue  = mode - nSigma * sigma
         highValue = mode + nSigma * sigma
         if(abs((sigma - testSigma) / sigma) .lt. tolerance) exit sigmaClip

         testSigma = sigma
      end do sigmaClip


      if(abs((sigma - originalSigma) / sigma) .gt. tolerance) then
         median = medianOfSorted(list(1:nList))
         mode = 2.5 * median - 1.5 * mean
      else
         mode = mean
      end if
      if (mode .le. 0.0 .or. mode .gt. 2.0 * median) mode = median

      deallocate(list)
   end function backgroundMode

end module omsourcebackground_mod
