module omsourcedetect_mod
   use cal
   use dal

   use omsourceutils_mod
   use omsourceparameter_mod
   use omsourcebackground_mod
   implicit none

   private ! all, but
   public  :: createLevelImage
   public  :: getCentrePixel

   character(*), parameter :: myModule="omsourcedetect_mod"

contains


   ! Info: creates an image that contains pixels with the source id as contents
   ! Originated: ComputeLevelImage
   function createLevelImage(image,delta,nSources,nSigma1) result(level)
      real(single),   pointer              :: image(:,:)
      integer(int32), intent(in)           :: delta
      integer(int32), intent(out)          :: nSources
      real(single),   intent(in), optional :: nSigma1
      integer(int16), pointer              :: level(:, :)

      real(single)   :: nSigma
      real(single)   :: sigma
      real(single)   :: mode
      real(single)   :: height
      integer(int32) :: i, j, k, l, m
      integer(int32) :: il, iu, jl, ju
      integer(int32) :: id
      integer        :: status
      integer(int32) :: lb(2),ub(2)
      integer(int32), parameter :: nIterations = 3    
      character(*), parameter :: routine=myModule//".createLevelImage"

      ! --- init output
      nullify(level)
      nSources=0
      if (.not.associated(image)) return ! don't bother

      ! --- init local
      lb=lbound(image)
      ub=ubound(image)
      id = 0

      if (present(nSigma1)) then
         nSigma = nSigma1
      else
         nSigma = 10.0
      end if


      allocate(level(lb(1):ub(1),lb(2):ub(2)), stat=status)
      call assertOn(status/=0,routine//".level")
      level = 0

      do k = 1, nIterations

         mode = computeBackgroundMode(image,level,nSigma, sigma)  
         height = mode + nSigma * sqrt(mode)
         do i = lb(1),ub(1)
            BLOCKJ : do j = lb(2),ub(2)

               if (image(i, j) > height .and. level(i, j) == 0) then
                  il = max(lb(1), i - delta)
                  iu = min(ub(1), i + delta)
                  jl = max(lb(2), j - delta)
                  ju = min(ub(2), j + delta)
                  do l = il, iu
                     do m = jl, ju
                        if (l == i .and. m == j) cycle
                        if (level(l, m) > 0) then
                           level(i, j) = level(l, m)

                           cycle BLOCKJ
                        end if
                     end do

                  end do

                  id = id + 1
                  level(i, j) = id                   
               end if

            end do BLOCKJ
         end do
      end do
      nSources = id 

      ! TODO: only for debugging
      if (.false.) call SaveLevelImage(level,"levelimage.fit", "level image")
   end function createLevelImage



   ! Info: Given an image, and given a map created by 'ComputeLevelImage'
   !       as above, this routine returns the pixel location of
   !       of 'id', that has the biggest image-value
   !       In other wodrs: find the centre pixel of source 'id'
   ! Limitations: image and level MUST be there, and have the same bounds
   ! Originated: ComputeSourcePosition
   function getCentrePixel(id,image,map) result(x)
      integer(int32), intent(in)  :: id ! of source
      real(single),   pointer     :: image(:,:)
      integer(int16), pointer     :: map(:,:)
      integer(int32)              :: x(2)

      integer(int32) :: i, j
      integer(int32) :: lb(2),ub(2)
      real(single)   :: value

      x=0
      if (.not.(associated(image) .and. associated(map))) return

      lb=lbound(image)
      ub=ubound(image)

      value = -1.e6 ! initial value
      do i = lb(1),ub(1)
         do j = lb(2),ub(2)
            if (map(i, j) == id) then
               if (image(i, j) > value) then
                  value = image(i, j)
                  x=(/i,j/)
               end if
            end if
         end do
      end do
   end function getCentrePixel



   ! Info: store the level image in fileName
   subroutine SaveLevelImage(level, fileName, heading)
      integer(int16), pointer    :: level(:,:)
      character(*),   intent(in) :: fileName
      character(*),   intent(in) :: heading

      type(DataSetT)             :: aSet 
      integer(int16), pointer    :: ptr(:,:)
      type(ArrayT)               :: anArray ! array for background

      if (.not.associated(level)) return

      aSet = DataSet(fileName, Create)
      anArray = addArray(aSet, heading, Integer16, shape(level))
      ptr => int16Array2Data(anArray)
      ptr = level

      call release(anArray)
      call release(aSet)
   end subroutine SaveLevelImage



end module omsourcedetect_mod
