! File:      omsourceparameter_mod
! Author:    Jorgo Bakker (jbakker@rssd.esa.int)
! Generated: Fri Aug  9 16:47:44 MEST 2002
! Usage:     C++/F95
! Example:
!   type(OmSourceParameter)          :: mySource
!   type(OmSourceParameter), pointer :: mySources(:)
!
!   call construct(mySource)         ! init contents
!   call new(mySources,low=0,high=5) ! allocate(s(0:5) and init...
!   :
!   call delete(mySources)
!   call destruct(mySource)
module omsourceparameter_mod
   use types
   use cal
   use errorHandling

   use omsourceutils_mod
   implicit none

   private  ! all, but (see for more information below)
   public  :: OmSourceParameter
   public  :: PhotometryParameters
   public :: ObsDetails   ! NB
   public :: ImageWindow
   public  :: construct
   public  :: destruct
   public  :: new
   public  :: delete
   public  :: setImageCoordProperties
   public  :: setMagnitudeProperties

   character(*), parameter :: myModule="omsourceparameter_mod"


    !*****************************************************
    ! Details about each source
    !*****************************************************

    type OmSourceParameter
        real(single)    :: xCoord              ! x coordinate (pix)
        real(single)    :: xCoordError         ! x coordinate error (pix)

        real(single)    :: yCoord              ! y coordinate pos (px)
        real(single)    :: yCoordError         ! y coordinate error (pix)
        real(double)    :: rightAscension
        real(double)    :: declination

        real(single)    :: positionError 

        real(single)    :: significance        ! significance
        real(single)    :: rawCountRate       ! raw count rate
        real(single)    :: rawCountRateError   ! raw count rate error

        real(single)    :: countRate           ! count rate
        real(single)    :: countRateError      ! count rate error
        real(single)    :: coincidenceCorrection  ! Coincidence loss correction
        real(single)    :: psf1Correction       ! PSF correction applied from r-12 pixels
        real(single)    :: psf2Correction     ! PSF correction for UV (12-35 pixels)

        real(single)    :: rawBackGroundCountRate ! background rate

        real(single)    :: backGroundCountRate ! background rate
        real(single)    :: magnitude           ! magnitude
        real(single)    :: exposure            ! Exposure (secs)
        real(single)    :: magnitudeError      ! magnitude error
        real(single)    :: calApertureRadius   ! cal aperture radius for doing photometry
        real(single)    :: photometryRadius    ! photometry radius
        real(single)    :: innerBackgroundRadius   ! Inner background radius
        real(single)    :: outerBackgroundRadius   ! Outer background radius

        real(single)    :: semiA               ! width of psf
        real(single)    :: semiB               ! width of psf  
        real(single)    :: paSource            ! position-angle 
        integer(int32)  :: noSourcePixels      ! number of pixels included by a source
        integer(int8)   :: extensionFlag       ! Point-source (0), otherwise extended (1)
        integer(int8)   :: confusionFlag       !
        integer(int8)   :: qualityFlag         ! Set to 1 if 1 or more source pixels is bad
    end type OmSourceParameter



    type PhotometryParameters
        real(single)    :: calApertureRadius              ! cal aperture radius (binned pixels)
    end type PhotometryParameters 
  

    !*****************************************************
    ! The observation details
    !*****************************************************
    type ObsDetails
        character(40) :: telescope
        character(40) :: observer
        character(40) :: object
        character(40) :: instrument
        character(40) :: obsID
        character(40) :: expID
        character(40) :: startObs
        character(40) :: endObs
        character(40) :: filter
        real(single)  :: exposure
    end type ObsDetails


    !*****************************************************
    ! The observation details
    !*****************************************************
    type ImageWindow
        integer(int32) :: x0
        integer(int32) :: y0
        real(double)   :: RA
        real(double)   :: Dec  
        real(double)   :: frameTime
        real(double)   :: deadFraction
        integer(int32) :: width
        integer(int32) :: height
        integer(int32) :: bx
        integer(int32) :: by
        logical        :: binBpe
        character(40)  :: filter
        character(40)  :: datamode
        real(double)   :: plateScale
    end type ImageWindow




!   interface construct
!      module procedure construct_od
!   end interface

!   interface destruct
!      module procedure destruct_od
!   end interface
  

   interface construct
      module procedure construct_
   end interface

   interface destruct
      module procedure destruct_
   end interface

   interface new
      module procedure new_
   end interface

   interface delete
      module procedure delete_
   end interface

   interface setImageCoordProperties
      module procedure setImageCoordProperties_
   end interface

   interface setMagnitudeProperties
      module procedure setMagnitudeProperties_
   end interface

contains

   subroutine construct_(s)
      type(OmSourceParameter),intent(out) :: s
      s%xCoord=0
      s%xCoordError=0.0_single
      s%yCoord=0.0_single
      s%yCoordError=0.0_single

      s%rightAscension=0.0_double
      s%declination=0.0_double

      s%positionError =0.0_single
      s%significance=0.0_single
      s%rawCountRate=0.0_single
      s%rawCountRateError=0.0_single
      s%countRate=0.0_single
      s%countRateError=0.0_single
      s%coincidenceCorrection=0.0_single
      s%psf1Correction=0.0_single
      s%psf2Correction=0.0_single
      s%rawBackGroundCountRate=0.0_single
      s%backGroundCountRate=0.0_single
      s%magnitude=0.0_single
      s%exposure=0.0_single
      s%magnitudeError=0.0_single
      s%calApertureRadius=0.0_single
      s%photometryRadius=0.0_single  ! was 0.0


      s%innerBackgroundRadius=0.0_single
      s%outerBackgroundRadius=0.0_single
      s%semiA=0.0_single
      s%semiB=0.0_single
      s%paSource=0.0_single

      s%noSourcePixels=0
      s%extensionFlag=0
      s%confusionFlag=0
      s%qualityFlag=0
   end subroutine construct_



   subroutine destruct_(s)
      type(OmSourceParameter),intent(inout) :: s
      ! we don't have to do something in here 
      ! but keep the compiler happy
      s%xCoord=0
   end subroutine destruct_



   subroutine new_(sources,low,high)
      type(OmSourceParameter),pointer    :: sources(:)
      integer,                intent(in) :: low,high

      integer                         :: i, status
      allocate(sources(low:high),stat=status)
      call assertOn(status/=0,myModule//".new.sources")
      do i=low,high
         call construct(sources(i))
      end do
   end subroutine new_



   subroutine delete_(sources)
      type(OmSourceParameter),pointer :: sources(:)

      if (associated(sources)) then
         deallocate(sources)
         nullify(sources)
      end if
   end subroutine delete_



   ! Info: sets x/y-coord + errors,and semi-axis
   ! Originated: SourceParameters_mod.ComputeSourceParameters
   ! TODO:       pointer to image, but fix computeMoments too!
   subroutine setImageCoordProperties_(s, image, backgroundMode, il, iu, jl, ju)
      type(OmSourceParameter),intent(inout) :: s      
!TODO      real(single),           pointer       :: image(:,:) ! (in)
      real(single),           intent(in)    :: image(:,:) ! (in)
      real(single) :: backgroundMode
      integer(int32),         intent(in)    :: il, iu, jl, ju


      real(single)   :: aValue, bValue, theta
      real(single)   :: xPos, yPos, x2, y2, xy
      real(single)   ::xPosErr, yPosErr
      real(single) :: diff, val1, val2
   
      ! ...
      call computeMoments(image, backgroundMode, il, iu, jl, ju, xPos, yPos, xPosErr, yPosErr, x2, y2, xy)

      ! Now convert these into major/minor values and theta
      ! factor of 2 somewhere?
      val1 = (x2 + y2) / 2.0
      val2 = sqrt(((x2 - y2) / 2.0) ** 2 + xy ** 2)
      aValue = sqrt(val1 + val2)

      if(val1 > val2) then
         bValue = sqrt(val1 - val2)
      else
         diff = abs(val1 - val2)

         if (diff > 1.0) then ! NB original JPDM criterion was 0
            ! There is a problem with the source parameterisation
            write(errstr,'(''Warning: Problem with source parameterisation - ignoring source'')')
            call message(AppMsg, Verbose)      ! Verbose was 2
            bValue=0.
         else

            bValue = aValue
         end if
      end if

      ! ... and position angle of the source:
      theta = getAngle(x2, y2, xy)

      ! --- now set the coordinate properties of the source
      s%xCoord=xPos
      s%xCoordError=xPosErr
      s%yCoord=yPos
      s%yCoordError=yPosErr

      s%semiA=aValue
      s%semiB=bValue
      s%paSource= theta * radiansToDegrees
   end subroutine setImageCoordProperties_


!*************************************************************************
! Info: computes the magnitude for a source, and sets the following
!       source properties:
!       - countRate, countRateError
!       - backgroundCountRate
!       - magnitude, magnitudeError
!       - psf1Correction, psf2Correction
!       - coincidenceCorrection
!       used properties:
!       - photometryRadius
!       - xCoord,yCoord
!       - rawCountRate, rawCountRateError
!       - rawBackgroundCountRate
! Originated: magnitude_mod.computeMagnitude
! Note: The source's photometryRadius property as well as maxRadius
!       are the binned versions TODO: hmm,did we want that?
!       If not, why are the (xCoord,yCoord) properties binned then?
!*************************************************************************
subroutine setMagnitudeProperties_(source, deadFrac, frameTime, maxRadius)
    type(OmSourceParameter), intent(inout) :: source
    real(double),            intent(in)    :: deadFrac
    real(double),            intent(in)    :: frameTime
    real(single),            intent(in)    :: maxRadius

    real(double) :: correctedTotalCountRate
    real(double) :: correctedBackgroundCountRate
    real(double) :: corrRate
    real(double) :: corrRateErrPlus  ! asymmetric
    real(double) :: corrRateErrMinus ! asymmetric
    real(double) :: corrRateErrPlus1 ! asymmetric
    real(double) :: corrRateErrMinus1! asymmetric
    real(double) :: corrRateErrPlus2 ! asymmetric
    real(double) :: corrRateErrMinus2! asymmetric
    real(double) :: corrMag          ! magnitude
    real(double) :: corrMagErrPlus   ! asymmetric
    real(double) :: corrMagErrMinus  ! asymmetric
    real(single) :: countFrameRatio
    real(double) :: totalCount
    real(double) :: countsRatio
    real(single) :: error, error1  
    logical      :: extend
      !character(*),parameter :: routine=myModule//".setMagnitudeProperties"

 
!write(*, *) "raw count rate=", source%rawCountRate
    if (source%rawCountRate .le. 0.0) then
        source%countRate             = -999.999
        source%countRateError        = -999.999
        source%backgroundCountRate   = -999.999
        source%magnitude             = -999.999
        source%magnitudeError        = -999.999
        ! TODO: mimic old behavior, but maybe you also want to set these:
        source%psf1Correction        = 0. ! TODO:-999.999
        source%psf2Correction        = 0. ! TODO:-999.999
        source%coincidenceCorrection = 0. ! TODO:-999.999
        return
    end if


    extend = .false.        
    source%countRate             = source%rawCountRate
    source%countRateError        = source%rawCountRateError
    source%backgroundCountRate   = source%rawBackgroundCountRate
    source%psf1Correction        = 0.0
    source%psf2Correction        = 0.0
    source%coincidenceCorrection = 0.0

    ! Scale counts to coincidence-loss correction radius
    countsRatio = 0.0
!write(*, *) "x=", source%xCoord, " y=", source%yCoord
!write(*, *) "source photometry radius=",source%photometryRadius, " calradius=", maxRadius 
    if (source%photometryRadius .lt. maxRadius) then
        ! Extrapolate counts from the omdetect aperture radius
        ! to 12-pixel radius and also the background counts

        ! Ensure ratio is between 0 and 1
        countFrameRatio = max( 0.0, min(real(source%countRate * frameTime), 1.0) )
        countsRatio = PSFCountsRatio(maxRadius, source%photometryRadius, &
            source%xCoord, source%yCoord, countFrameRatio)

        source%countRate = source%countRate * countsRatio
        source%countRateError = source%countRateError * countsRatio
        source%backGroundCountRate = source%backGroundCountRate * &
              (maxRadius / source%photometryRadius) ** 2 

        source%psf1Correction = (countsRatio - 1.0) * 100.0
    end if
!write(*, *) "psf1 correction=",   source%psf1Correction  
!write(*, *) "exposure =", source%exposure
    ! Need the total count rate (background+source) for the coincidence-loss correction
    totalCount = dble(source%countRate + source%backGroundCountRate)

    ! Get the new count rate by subtracting the corrected background 
    ! count-rate from the corrected total count-rate

    ! No error is stored for the background so use sqrt of counts
    !  error : in the background rate
    !  error1: in the total count rat
    error  = sqrt(source%backGroundCountRate / source%exposure)
    error1 = sqrt(error**2 + source%countRateError**2)

    ! Correct the total count rate
    correctedTotalCountRate = &
           CorrectedCountRate(totalCount, error1, frameTime, deadFrac, &
           extend, corrRateErrMinus1, corrRateErrPlus1)

    ! Correct the source and background count rate
    correctedBackgroundCountRate = CorrectedCountRate(dble(source%backGroundCountRate), error, frameTime, &
           deadFrac, extend,  corrRateErrMinus2, corrRateErrPlus2) 
    source%backgroundCountRate = correctedBackgroundCountRate 

    corrRate =  correctedTotalCountRate -  correctedBackgroundCountRate 
    source%countRate = corrRate 

    if (source%rawCountRate .gt.0.0) then
        source%coincidenceCorrection = (source%countRate - source%rawCountRate) * 100.0 /  source%rawCountRate 
    end if
  
    if (corrRate .gt. 0.0) then
        corrRateErrMinus = sqrt(corrRateErrMinus1**2 + corrRateErrMinus2**2)
        corrRateErrPlus  = sqrt(corrRateErrPlus1**2  + corrRateErrPlus2**2 )

        ! If necessary, scale the counts to the CAL photometric radius
        countsRatio = 0.0
        if (source%photometryRadius .gt. maxRadius) then
            ! Ensure ratio is between 0 and 1
            countFrameRatio = max(0.0, min(real(corrRate * frameTime), 1.0) )
            countsRatio = PSFCountsRatio(maxRadius, source%photometryRadius, &
                 source%xCoord, source%yCoord, countFrameRatio)

            corrRate         = corrRate * countsRatio
            corrRateErrPlus  = corrRateErrPlus * countsRatio
            corrRateErrMinus = corrRateErrMinus * countsRatio

            source%CountRate           = source%countRate * countsRatio
            source%CountRateError      = source%countRateError * countsRatio
            source%BackgroundCountRate = source%backgroundCountRate * countsRatio
            source%psf2Correction      = (countsRatio - 1.0) * 100.0
        end if

        call CAL_omPhotoMagnitude(corrRate, corrMag, -corrRateErrMinus, &
              corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)

        source%magnitude      = real(corrMag)
        source%magnitudeError = real(abs(max(corrMagErrMinus, corrMagErrPlus)))
    end if


end subroutine setMagnitudeProperties_


!********************************************************************************
! private helper functions
! --------------------------------------------------------------------------

! Originated: SourceParameters_mod.ComputeMoments
! Purpose: Loops through the number of sources found for each 
! source in the crude source list (this value will normally be 1
! and hence this routine will usually be called once for each 
! of the initially detected sources)
! Each source is parameterised using moments
! Returns the x and y positions, countrates 
! and background rates for each source
! Parameters:
! image - A 3-d array- one 2-d image for each source
!***************************************************************************

subroutine computeMoments(image,  backGround, il, iu, jl, ju, &
        xPos, yPos, xPosErr, yPosErr, &
        x2, y2, xy)
    integer(int32),intent(in)   :: il, iu, jl, ju
!TODO      real(single),           pointer       :: image(:,:) ! (in)
    real(single),           intent(in)    :: image(:,:) ! (in)
    real(single), intent(in) :: backGround
    real(single),  intent(out)  :: xPos
    real(single),  intent(out)  :: yPos
    real(single),  intent(out)  :: xPosErr
    real(single),  intent(out)  :: yPosErr
    real(single),  intent(out)  :: x2
    real(single),  intent(out)  :: y2
    real(single),  intent(out)  :: xy

    integer(int32)              :: i, j
    real(double)                :: imageVal
    real(double)                :: xErrTot, yErrTot
    real(double)                :: xposD, yposD, x2D, y2D, xyD
    real(double)                :: totalImage, totalImage2    
    real(single),  parameter    :: pixelIgnoreFlag = -1e8


    totalImage = 0.0_double
    totalImage2 = 0.0_double

    xposD = 0.0_double
    yposD = 0.0_double
    x2D = 0.0_double
    y2D = 0.0_double
    xyD = 0.0_double

    ! Compute the x and y centroid positions (xpos & ypos)
    do i = il,  iu
        do j = jl, ju
           
            imageVal = dble(image(i, j)) - backGround ! NB - The image has not been background subtracted

            if(imageVal .ne. pixelIgnoreFlag) then
                xposD = xposD + imageVal * i
                yposD = yposD + imageVal * j
                totalImage2 = totalImage2 + imageVal * imageVal
                totalImage = totalImage + imageVal

            end if
        end do
    end do

    xErrTot = 0.0_double
    yErrTot = 0.0_double
    xposD = xposD / totalImage
    yposD = yposD / totalImage
    xpos = real(xposD) - 0.5
    ypos = real(yposD) - 0.5
    do i = il, iu

        do j = jl, ju
            imageVal = dble(image(i, j))
            if(imageVal .ne. pixelIgnoreFlag) then

                !*******************************************************
                ! Assume that the error is dominated by the total counts
                !*******************************************************
                xErrTot = xErrTot + imageVal * dble((j - xposD) ** 2)
                yErrTot = yErrTot + imageVal * dble((i - yposD) ** 2)
               
   

                x2D = x2D + imageVal * dble(i * i)
                y2D = y2D + imageVal * dble(j * j)
                xyD = xyD + imageVal * dble(i * j)
            end if
        end do
    end do

    xPosErr = sqrt(xErrTot / totalImage2)   ! Source pixel x-position error
    yPosErr = sqrt(yErrTot / totalImage2)   ! Source pixel y-position error

    x2 = real(x2D / totalImage - xposD * xposD)
    y2 = real(y2D / totalImage - yposD * yposD)
    xy = real(xyD / totalImage - xposD * yposD)
end subroutine computeMoments



function getAngle(x, y, xy) result(angle)
      real(single), intent(in) :: x, y, xy
      real(single)             :: angle

      real(single)             :: theta1, theta2 

      if(x .ne. y) then
         theta1 = atan(2 * xy / (x - y)) / 2.0
         theta2 = theta1 + HALF_PI
         if(theta2 .gt. HALF_PI) theta2 = theta2 - PI
         if(xy .ge. 0.0) then
            if(theta1 .ge. 0.0) then
               Angle = theta1
            else
               Angle = theta2
            end if
         else
            if(theta1 .lt. 0.0) then
               Angle = theta1
            else
               Angle = theta2
            end if
         end if
      else
         Angle = sign(real(HALF_PI), xy) / 2.0
      end if
   end function getAngle




!**********************************************************************
! Info: Corrects the raw count rate for coincidence losses, etc  
! Originated: magnitude_mod.CorrectedCountRate
!*********************************************************************
function CorrectedCountRate(rawCountRate, & 
      countRateError, frameTime, deadFrac, extended, &
        countRateErrorMinus, countRateErrorPlus)
    real(double), intent(in)    :: rawCountRate
    real(single), intent(in)    :: countRateError
    real(double), intent(out)   :: countRateErrorMinus
    real(double), intent(out)   :: countRateErrorPlus
    real(double), intent(in)    :: frameTime
    real(double), intent(in)    :: deadFrac
    logical,      intent(in)    :: extended
    real(double)                :: CorrectedCountRate

    logical                     :: isExtended
    type(ValueAsymmetricErrorT) :: rawRate
    type(ValueAsymmetricErrorT) :: correctedRate

    isExtended=extended ! TODO: dummy statement;remove argument?
    if (rawCountRate > (1.0_double / frameTime)) then
         isExtended = .true.
        write(errstr,*) &
              "Count-rate>frame-rate,Coincidence-loss correction ignored",&
              "rawcountrate=", rawCountRate, " Frame time=", frameTime 
        call warning("CountRateTooHigh")
    else
         isExtended = .false.
    end if

    rawRate = ValueAsymmetricErrorT(rawCountRate, dble(countRateError), dble(countRateError))
    correctedRate = CAL_omPhotoNatural(rawRate, frameTime, deadFrac, 0.0_double, isExtended, .TRUE.)

    ! If the count rate is greater than the frame rate then the above
    ! routine cannot do the correction- instead it sets the corrected 
    ! count rate to the largest number that can be stored on the computer. 
    if (CorrectedRate%value .lt. 1e10) then
        CorrectedCountRate = CorrectedRate%value
        countRateErrorMinus = correctedRate%errorMinus
        countRateErrorPlus  = correctedRate%errorPlus
    else
        CorrectedCountRate = -1.0d0
    end if
end function CorrectedCountRate


!************************************************************
! Info: Computes the ratio of the total counts between
!       the radii radius1 and radius2 
! Originated: magnitude_mod.PSFCountsRatio
!************************************************************
Function PSFCountsRatio(radius1, radius2, x0, y0, countFrameRatio)
    real(single), intent(in) :: radius1    ! The aperture radius that the input counts refer to
    real(single), intent(in) :: radius2    ! The radius that the counts have to be scaled to 
    real(single), intent(in) :: x0         ! The x coordinate of the source 
    real(single), intent(in) :: y0         ! The y coordinate of the source 
    real(single), intent(in) :: countFrameRatio
    real(double)             :: PSFCountsRatio

    PSFCountsRatio = PSFFractionOfTotalCounts(x0, y0, radius1, countFrameRatio) / &
           PSFFractionOfTotalCounts(x0, y0, radius2, countFrameRatio)
end function PSFCountsRatio


!*************************************************************
! Info: Returns the fraction of the total counts
!       contained within the given radius
! Originated: magnitude_mod.PSFFractionOfTotalCounts
!*************************************************************
function PSFFractionOfTotalCounts(x0, y0, radius, countFrameRatio)
    real(single), intent(in) :: x0
    real(single), intent(in) :: y0
    real(single), intent(in) :: radius
    real(single), intent(in) :: countFrameRatio
    real(double)             :: PSFFractionOfTotalCounts

    type(OmPsfT) :: psf

    psf = CAL_omGetPSF(dble(x0), dble(y0), dble(countFrameRatio))
    PSFFractionOfTotalCounts = CAL_omPsfEncircledEnergy(psf, dble(radius))
end function PSFFractionOfTotalCounts


end module omsourceparameter_mod
