! module omsourcephotometry_mod (10th July 2001)
!
! Module containing routines, etc,
! for doing aperture photometry on each source. I has been added
! because the counts that initially computed for each source (at the same
! time that the centroided positions for each source are obtained) is not
! consistent with the CAL calibration of source counts to magnitudes.
! It also sets the source confusion flag and the source
! quality flag.
module omsourcephotometry_mod
   use errorhandling
   use cal
   use dal

   use omsourceutils_mod
   use omsourcebackground_mod
   use omsourceparameter_mod
   implicit none

   private ! all, but:
   public  :: createIdMap
   public  :: aperturePhotometry

   character(*), parameter :: myModule="omsourcephotometry_mod"

   type point
      real(single) :: x
      real(single) :: y
      real(single) :: radius
   end type point

contains
!**********************************************************************************
!
! Performs aperture photometry on a set of sources using a circular aperture
! of a given pixel radius
! works on array of OmSourceParameter's:
! -in/out:
!  src%photometryRadius
!  src%semiA
!  src%semiB
!  src%rawCountRate
!  src%rawCountRateError
!  src%significance
!  src%rawBackgroundCountRate
! -out:
!  src%confusionFlag
!  src%qualityFlag
subroutine AperturePhotometry(image, quality, sources, apertureDefault, fixedAp)
    real(single),            pointer       :: image(:,:)
    logical(bool),           pointer       :: quality(:,:)
    type(OmSourceParameter), intent(inout) :: sources(:)
    real(single),            intent(in)    :: apertureDefault
    logical,optional,        intent(in)    :: fixedAp

    ! Local variables
    logical        :: hasFixedAperture  ! If TRUE then the aperture is fixed on input
    real(double)   :: sourceCounts      ! The aperture counts for a source
    real(double)   :: backGroundCounts  ! The background counts for a source
    real(double)   :: backGroundCountsPerPixel
    real(single)   :: apertureRadiusMin ! The minimum allowed
    real(double)   :: apertureArea     ! actual aperture area used to get counts
    real(single)   :: fullApertureArea ! required aperture area used to get counts
    real(single)   :: maxRadius
    real(single)   :: backGroundCountsError
    real(single)   :: bgrMode   ! The background mode value
    real(single)   :: sigma
    integer(int32) :: idx,nSources
    integer(int32) :: i, k, status, apertureType
    real(single)   :: xLeft,yBottom,xRight,yTop
    integer(int32)          :: nBadPixels
    integer(int32), pointer :: sortedIds(:)
    integer(int16), pointer :: map(:,:)
    real(single),   pointer :: bgrPixels(:)
    type(point),    pointer :: sourcePoints(:)
    real(single),   pointer :: objectWidths(:) ! maximum fwhm of each source


    ! Note: Set minSignificance to zero if no counts extrapolation
    ! to be used ! NB until the CAL calibration data is in place
    ! the full aperture will be used on all sources, unless
    ! another source lies within the aperture. If the source
    ! significance is less than this value then
    ! the photometric counts will be extrapolated using
    ! the point-spread-function
    real(single),  parameter :: minSignificance = 0.0
    character(*) , parameter :: routine=myModule//".AperturePhotometry"

    nSources=size(sources)
    if (nSources < 1) return ! don't bother

    write(errstr,'(a)') "Performing aperture photometry on the sources"
    call message(AppMsg, Verbose)

    if (present(fixedAp)) then
        hasFixedAperture=fixedAp
    else
        hasFixedAperture=.false.
    end if


      ! --- Allocate memory for arrays and check if successful
    map=>createIdMap(image,sources,apertureDefault)

    allocate(sourcePoints(nSources), stat=status)
    call assertOn(status/=0,routine//".sourcePoints")

    allocate(objectWidths(nSources), stat=status)
    call assertOn(status/=0,routine//".objectWidths")


    ! Set the minimum aperture radius to half the required aperture
    apertureRadiusMin = apertureDefault / 2.
    !fullApertureArea = PI *sources%photometryRadius*sources%photometryRadius

    bgrPixels=>createBackGroundPixel(&
           minval(sources%innerBackgroundRadius), &
           maxval(sources%outerBackgroundRadius))


    ! Obtain an index of the source positions in
    ! increasing x-pixel position. NB sources%xCoord is not
    ! sorted. sortedIds(i) will give the index of the
    ! array element that gives the positions in increasing
    ! x-position. This is used for obtaining the distance
    ! from one source to its nearest neighbour.
    allocate(sortedIds(nSources),stat=status)
    call assertOn(status/=0,routine//".sortedIds")
    sortedIds = sortedIndexOf(sources%xCoord)


    ! Compute the mode of the image, ignoring source regions
    ! Multiply it by 3- when getting background ignore pixel
    ! values greater than this value
    bgrMode = computeBackgroundMode(image,map,3.0,sigma)


    ! sourcePoints will contain the position of the sources
    ! in increasing x-position- this array is used to obtain
    ! the maximum-sized aperture for doing photometry.




    do i = 1, nSources
        idx = sortedIds(i)

        !  TODO: (i) should it not be (idx) ???
        sourcePoints(i)%x = sources(idx)%xCoord
        sourcePoints(i)%y = sources(idx)%yCoord

        ! objectWidths will store the maximum of the x and y extents of the
        ! sources* 3, or if this is less than the aperture radius, to the
        ! aperture radius.
        objectWidths(i) = 3.0 * max(sources(idx)%semiA, sources(idx)%semiB)
        objectWidths(i) = sources(i)%photometryRadius !TODO: sources(idx)?
    
!write(*, *) "omsourcephotometry line 142 after sorting source(1) = ",sources(1)%xCoord, " ", sources(1)%yCoord 
    end do

      ! --- now do aperture photometry on each source, if possible

    xLeft   = lbound(image,1)
    yBottom = lbound(image,2)
    xRight  = ubound(image,1)
    yTop    = ubound(image,2)

    k = 0
    do i = 1,  nSources
        idx = sortedIds(i)
        sources(idx)%qualityFlag=0

        ! For the moment do photometry on extended sources
        if (sources(idx)%extensionFlag .lt. 10) then
!write(*, *) "input  no " , i, "  ", sources(i)

            ! Find the aperture radius that can be used
            ! for the photometry. If type is 1 then
            ! can use the required aperture, 2 - have
            ! to use a smaller one, 3 can't do photometry
            fullApertureArea = PI * sources(idx)%photometryRadius**2
            apertureArea=0.0_double

            if (hasFixedAperture) then
                maxRadius = sources(idx)%photometryRadius
                apertureType = 1
            else
                maxRadius = GetMaximumAperture(sourcePoints, objectWidths, i, &
                  xLeft, yBottom, xRight, yTop, apertureRadiusMin, &
                  sources(idx)%photometryRadius, apertureType)
            end if

            select case (apertureType)
                case (1, 2, 4)
                    ! 1 - No other objects within aperture,
                    ! 2 - Another object within aperture, but outside
                    !     minimum aperture radius- use aperture that
                    !     excludes this source
                    ! 4 - ???

              
                if (sources(idx)%significance .lt. minSignificance) then
                    ! Use a smaller aperture
                    maxRadius = apertureRadiusMin
                end if

                call getApertureCounts(image, quality, maxRadius, sources(idx), apertureArea, sourceCounts, nBadPixels)
                if (nBadPixels .gt. 0) then
                    ! Bad pixels were found: set source quality flag
                    sources(idx)%qualityFlag=1
                end if

                backGroundCountsPerPixel = &
                    GetBackGroundCounts(image,quality,map,bgrPixels,&
                    sources(idx), bgrMode,backgroundCountsError)

                backGroundCounts     = backGroundCountsPerPixel * apertureArea
                backGroundCountsError= backGroundCountsError * apertureArea
                sourceCounts         = sourceCounts - backGroundCounts


                if (apertureType .eq. 4) then
                    sourceCounts = sourceCounts * fullApertureArea / apertureArea
                    maxRadius = sources(idx)%photometryRadius
                    backGroundCounts = backGroundCounts * fullApertureArea / apertureArea
                    backGroundCountsError = backGroundCountsError * fullApertureArea / apertureArea
                end if

                sources(idx)%photometryRadius = maxRadius

                ! Recalculate the source significance
                sources(idx)%significance = &
                    ComputeSourceSignificance(sourceCounts, backGroundCounts)



                ! Replace the old source counts with the new value
                sources(idx)%rawCountRate = sourceCounts
                sources(idx)%rawCountRateError = &
                    ComputeSourceCountsError(sourceCounts, backGroundCounts)
                sources(idx)%rawBackgroundCountRate = backGroundCounts


                case (3) ! Object within aperture radius/2.0
               ! leave as original for now
            end select

        else
            apertureType = 1
        end if


    ! --- Set the source confusion flag
    ! 0 isolated source
    ! 1 apertureRadiusMin < sourceRadius < aperture radius
    ! 2 sourceRadius < apertureRadiusMin
    sources(idx)%confusionFlag = apertureType

!write(*, *) "output no " , i, "  ", sources(idx)%xCoord, " ", sources(idx)%yCoord
!write(*, *) "source counts=", sourceCounts, " background counts=", backGroundCounts 

    end do

      ! clean-up:
    deallocate(bgrPixels)
    deallocate(map)
    deallocate(sortedIds)
    deallocate(sourcePoints)
    deallocate(objectWidths)

    nullify(bgrPixels)
    nullify(map)
    nullify(sortedIds)
    nullify(sourcePoints)
    nullify(objectWidths)
end subroutine AperturePhotometry


!**********************************************************
! Info: Computes the counts within a given aperture.
!       Returns the value via the function name.
! Returns:
!       nBad - number of bad pixels ignored
!       area - weighted sum of all pixel areas used
!       counts - sum of pixel contents for this area
!***********************************************************
subroutine getApertureCounts(image,quality,radius,source,area,counts,nBad)
    real(single),            pointer       :: image(:,:)
    logical(bool),           pointer       :: quality(:,:)
    real(single),            intent(in)    :: radius ! of aperture
    type(OmSourceParameter), intent(in)    :: source
    integer(int32),          intent(out)   :: nBad
    real(double),            intent(out)   :: area   ! actual area used
    real(double),            intent(out)   :: counts !

      ! Local variables
    real(double) :: pArea
    integer      :: i, j, i0, j0, r
    integer      :: i1, i2, j1, j2
    integer      :: il,iu,jl,ju

    counts = 0.0_double
    area   = 0.0_double
    nBad   = 0
    r = ceiling(radius)

    il = lbound(image,1)
    iu = ubound(image,1)
    jl = lbound(image,2)
    ju = ubound(image,2)

    i0 = int(source%xCoord)
    j0 = int(source%yCoord)

    ! Determine upper and lower array indices for
    ! summing counts
    i1 = i0 - r - 1
    i1 = max(il, i1)

    i2 = i0 + r + 1
    i2 = min(iu, i2)

    j1 = j0 - r - 1
    j1 = max(jl, j1)

    j2 = j0 + r + 1
    j2 = min(ju, j2)

    do i = i1,i2
        do j = j1,j2
            if (quality(i, j)) then
               nBad = nBad + 1
               cycle
            end if

            ! how much of the belongs to aperture?
            pArea   = pixelArea(i,j,radius,source)
            if (pArea == 0.0_double) cycle

            counts= counts + image(i,j) * pArea
            area  = area + pArea
        end do
    end do
end subroutine getApertureCounts


!**********************************************************************
! Info: returns the fraction of area for pixel(i,j) that belongs to
!       the circle(source%xCoord,source%yCoord,r)
! Note: 0 <= area <=1
!*********************************************************************
function pixelArea(i, j, r, source) result(area)
    integer(int32),          intent(in)    :: i,j ! image location
    real(single),            intent(in)    :: r   ! aperture
    type(OmSourceParameter), intent(in)    :: source
    real(double)                           :: area

    real(single) :: r1, r2, r3, r4
    real(single) :: x, y, x1, x2, y1, y2
    real(single) :: minRadius, maxRadius


    x = real(i)
    y = real(j)

    x1 = x - source%xCoord
    y1 = y - source%yCoord
    x2 = x1 + 1.0
    y2 = y1 + 1.0

    r1 = sqrt(x1 * x1 + y1 * y1)
    minRadius = r1
    maxRadius = r1

    r2 = sqrt(x2 * x2 + y1 * y1)
    minRadius = min(minRadius, r2)
    maxRadius = max(maxRadius, r2)

    r3 = sqrt(x1 * x1 + y2 * y2)
    minRadius = min(minRadius, r3)
    maxRadius = max(maxRadius, r3)

    r4 = sqrt(x2 * x2 + y2 * y2)
    minRadius = min(minRadius, r4)
    maxRadius = max(maxRadius, r4)


    if (r .le. minRadius) then
        area=0.0_double
    elseif (r  .ge. maxRadius) then
        area=1.0_double
    else
        area=effectivePixelArea(x1,y1,r)
    end if
end function pixelArea


!**********************************************************************
! Info: Computes the area of a pixel contained within the aperture and
! Note: This will shortly be replaced by a better routine
!**********************************************************************
function effectivePixelArea(x1, y1, r) result(area)
    real(single), intent(in)  :: x1, y1,r
    real(double)              :: area

    real(single)              :: radiusSqr
    real(single)              :: xPos, yPos
    integer(int32)            :: nIn
    integer(int32)            :: i, j
    real(single)              :: rsq
    integer(int32), parameter :: maxTimes = 99 * 99

    rsq = r**2
    nIn = 0

    do i = 2, 100

        xPos = x1 + 0.01 * real(i - 1)
        do j = 2, 100
            yPos = y1 + 0.01 * real(j - 1)

            radiusSqr = xPos * xPos + yPos * yPos
            if (radiusSqr .le. rsq) then
               nIn = nIn + 1
            end if
        end do
    end do

    area = dble(nin) / dble(maxTimes)
end function effectivePixelArea


!************************************************
! Info: Computes the source counts error
! Args:
!       src - The background-subtracted source counts
!       bgr - The background counts
function ComputeSourceCountsError(src,bgr) result(e)
    real(double), intent(in) :: src
    real(double), intent(in) :: bgr
    real(double)             :: e

    e = sqrt(src + 2.0_double * bgr)
end function ComputeSourceCountsError


!********************************************************
! Info: Computes the source significance
! Args:
!       src - The background-subtracted source counts
!       bgr - The background counts
!*******************************************************
function ComputeSourceSignificance(src, bgr) result(s)
    real(double), intent(in) :: src
    real(double), intent(in) :: bgr
    real(double)             :: s

    s = src / sqrt(bgr)
end function ComputeSourceSignificance


!*************************************************************************
! Info: Finds the maximum aperture size that can be used for photomery
! Side effect: apId (aperture type):
!       -1 = no object within aperture radius pixels
!       -2 = object within aperture radius, but outside minimum aperture radius
!       -3 = object within minimum aperture radius
!       -4 = ??? TODO:describe
!**************************************************************************
function GetMaximumAperture(sourcePoints, objectWidths, id, xLeft, yBottom, xRight, yTop, rMin, rAp, apId) result(radius)
    type(point),    pointer     :: sourcePoints(:)
    real(single),   pointer     :: objectWidths(:)
    integer(int32), intent(in)  :: id       ! Index of source
    real(single),   intent(in)  :: xLeft,yBottom,xRight,yTop
    real(single),   intent(in)  :: rMin ! The minimum allowed radius
    real(single),   intent(in)  :: rAp  ! The current aperture radius
    integer(int32), intent(out) :: apId! Determines what kind of photometry can be performed
    real(single)                :: radius

    type(point)    :: centre             ! Coords of source
    real(single)   :: minimumSeparation  ! Minimum distance between source of interest and other sources
    real(single)   :: borderDistance     ! Distance to the edge of the image from the source
    real(single)   :: okayDistance       ! Minimum allowed separation between source and nearest neighbour
    real(single)   :: distance           ! Distance bewteen source and neighbour
    real(single)   :: d, d1
    integer(int32) :: k
    integer(int32) :: nSources   ! The number of sources

      nSources=size(sourcePoints) ! = size(objectWidths)
      centre = sourcePoints(id)

      okayDistance = rAp
      borderDistance = 1e8
      minimumSeparation = 1e6


      ! Check to see if object is near the edge of the image
      if (centre%x - xLeft .lt. rAp) then
         borderDistance = centre%x - xLeft
         apId = 4
      else if  (xRight - centre%x .lt. rAp) then
         apId = 4
         borderDistance = xRight - centre%x
      end if

      if (centre%y - yBottom .lt. rAp) then
         apId = 4
         borderDistance = min(centre%y - yBottom, borderDistance)
      else  if (yTop - centre%y .lt. rAp) then
         apId = 4
         borderDistance = min(yTop - centre%y, borderDistance)
      end if


      ! Find minimum distance between source of interest and sources to its
      ! left.
      ! Note: sourcePoints contains the x-coordinates of the sources in
      !       increasing order.
      if (id .gt. 1) then
         k = id - 1

         do while (k .ge. 1)
            d = centre%x - sourcePoints(k)%x
            d = max(0.0, d)
            if (d .gt. rAp + objectWidths(k) ) exit
            if (sourcePoints(k)%y .ge. centre%y) then
               d1 = sourcePoints(k)%y - centre%y
               d1 = max(0.0, d1)

            else
               d1 = centre%y - sourcePoints(k)%y
               d1 = max(0.0, d1)
            end if

            distance = sqrt(d * d + d1 * d1)
            if (distance + rAp - objectWidths(k) .lt. minimumSeparation) then

               minimumSeparation = distance + rAp - objectWidths(k)

            end if


            k = k - 1
         end do
      end if

      ! Find minimum distance between source of interest
      ! and sources to its right.
      if (id .lt. nSources) then
         k = id + 1
         do while (k .le. nSources)
            d = sourcePoints(k)%x - centre%x
            d = max(0.0, d)

            if (d .gt. rAp + objectWidths(k)) exit

            if (sourcePoints(k)%y .ge. centre%y) then
               d1 = sourcePoints(k)%y - centre%y
               d1 = max(0.0, d1)
            else
               d1 = centre%y - sourcePoints(k)%y
               d1 = max(0.0, d1)
            end if

            distance = sqrt(d * d + d1 * d1)
            if (distance + rAp - objectWidths(k) .lt. minimumSeparation) then
               minimumSeparation = distance + rAp - objectWidths(k)
            end if
            k = k + 1
         end do
      end if

      if (minimumSeparation .ge. okayDistance) then

         if (apId .ne. 4) apId = 1
         radius = rAp

      else if (minimumSeparation .ge. rMin) then
         apId = 2
         radius = minimumSeparation   ! - rAp

      else
         apId = 3    ! Can't do photometry on
         radius = 0.0
      end if

end function GetMaximumAperture


!***************************************************************************
! Info: Computes the background counts within the given annulus, using the
!       median counts/pixels obtained from an annulus region.
! Side effect: error (standard deviation)
!***************************************************************************
function GetBackGroundCounts(image, quality, map, bgPix, source, defBgCounts, error) result(counts)
    real(single),            pointer    :: image(:,:)
    logical(bool),           pointer    :: quality(:,:)
    integer(int16),          pointer    :: map(:,:)
    real(single),            pointer    :: bgPix(:)
    type(OmSourceParameter), intent(in) :: source
    real(single),            intent(in) :: defBgCounts
    real(single),            intent(out):: error
    real(double)                        :: counts

    ! Local variables
    integer(int32)  :: i, j
    integer(int32)  :: i0, j0, i1, i2, j1, j2
    integer(int32)  :: ndata  ! The number of background pixels used
    real(single)    :: x, y, dx, dy, r
    integer        :: il,iu,jl,ju

    ! Scan the annulus region and add pixel values to the
    ! bgPix vector
    i0 = int(source%xCoord)
    j0 = int(source%yCoord)
    il = lbound(image,1)
    iu = ubound(image,1)
    jl = lbound(image,2)
    ju = ubound(image,2)
    r  = ceiling(source%outerBackGroundRadius)
    nData = 0

    ! Determine upper and lower array indices for
    ! summing counts
    i1 = i0 - r - 1
    i1 = max(il, i1)

    i2 = i0 + r + 1
    i2 = min(iu, i2)

    j1 = j0 - r - 1
    j1 = max(jl, j1)

    j2 = j0 + r + 1
    j2 = min(ju, j2)

    do i = i1, i2

        x = real(i)
        dx = x - source%xCoord

        do j = j1, j2
            if (.not.quality(i,j) .and. map(i,j)==0) then ! Pixel okay to use
                y = real(j)
                dy = y - source%yCoord
                r = sqrt(dx * dx + dy * dy)
                if (r>=source%innerBackGroundRadius .and. r<=source%outerBackGroundRadius) then
                    nData = nData + 1
                    bgPix(nData) = image(i, j)
                end if
            end if
        end do
    end do

    ! Compute the median background pixel value
    if (nData > 0) then
        counts = computeBackgroundMode1(bgPix(1 : nData), 2.0, error)
    else
        call warning("NoBackgroundMode","No data points for background")
        counts = defBgCounts
        error  = 0.0
    end if

end function GetBackgroundCounts

   ! Info: Computes the fraction of total counts within the given radius
   function PSFFractionOfTotalCounts(x0,y0,radius,ratio) result(counts)
      real(double), intent(in) :: x0
      real(double), intent(in) :: y0
      real(double), intent(in) :: radius
      real(double), intent(in) :: ratio ! count frame ratio
      real(double)             :: counts

      type(OmPsfT) :: psf

      psf = CAL_omGetPSF(x0, y0, ratio)
      counts = CAL_omPsfEncircledEnergy(psf, radius)
   end function PSFFractionOfTotalCounts



   ! Info: Computes the radius containing the fraction of the total counts
   function PSFRadiusContainingCounts(x0,y0,fraction,ratio) result(radius)
      real(double), intent(in) :: x0
      real(double), intent(in) :: y0
      real(double), intent(in) :: fraction
      real(double), intent(in) :: ratio ! count frame ratio
      real(double)             :: radius

      type(OmPsfT) :: psf

      psf = CAL_omGetPSF(x0, y0, ratio)
      radius = CAL_omPsfCircleRadius(psf, fraction)
   end function PSFRadiusContainingCounts



   ! returns an image. All pixels that belong to the same source have
   ! the same contents (source id)
   function createIdMap(image,sources,radiusIn) result(ids)
      real(single),            pointer    :: image(:,:) ! (in)
      type(OmSourceParameter), intent(in) :: sources(:)
      real(single),            intent(in) :: radiusIn
      integer(int16),          pointer    :: ids(:,:)

      real(single)   :: r, radius,  x0, y0, xPixel, yPixel
      integer        :: i, j, k, i0, j0, i1, i2, j1, j2,status
      integer        :: il,iu,jl,ju
      integer(int32) :: nSources
      character(*) , parameter :: routine=myModule//".createIdMap"

      nSources=size(sources)
      r = ceiling(radiusIn)

      il = lbound(image,1)
      iu = ubound(image,1)
      jl = lbound(image,2)
      ju = ubound(image,2)

      allocate(ids(il:iu,jl:ju), stat=status)
      call assertOn(status/=0,routine//".ids")
      ids = 0

      do k = 1, nSources
         x0 = sources(k)%xCoord
         y0 = sources(k)%yCoord
         i0 = int(x0)
         j0 = int(y0)

         ! Determine upper and lower array indices for summing counts
         i1 = i0 - r - 1
         i1 = max(il, i1)

         i2 = i0 + r + 1
         i2 = min(iu, i2)

         j1 = j0 - r - 1
         j1 = max(jl, j1)

         j2 = j0 + r + 1
         j2 = min(ju, j2)

         do i = i1, i2
            xPixel = int(i)
            do j = j1, j2
               yPixel = int(j)
               radius = sqrt((xPixel - x0) ** 2 + (yPixel - y0) ** 2)

               if (radius .le. radiusIn) then
                  ids(i, j) = k
               end if

            end do
         end do

      end do
   end function createIdMap


   ! Info: allocates memory for bgPix and bgOff
   function createBackgroundPixel(r0,r1) result(bgPix)
      real(single), intent(in) :: r0, r1
      real(single), pointer    :: bgPix(:)

      integer(int32)           :: nNeeded, status
      character(*), parameter  :: routine=myModule//".createBackgroundPixel"

      nNeeded = ceiling(PI * (r1**2 - r0**2)) + 50 !TODO: why the extra 50?

      allocate(bgPix(nNeeded), stat=status)
      call assertOn(status/=0,routine//".bgPix")
      bgPix=0.
   end function createBackgroundPixel

end module omsourcephotometry_mod
