! -*-f90-*-
!
! Author:    jbakker
! Generated: Mon Sep 23 13:17:55 CEST 2002
! Usage:     -
! Info:      This is a **temporary** module and acts as a placeholder
!            until all SAS fortran developers have sorted out their act
!            together
module omsourceutils_mod
   use types
   use errorHandling, only : fatal
   implicit none

   public ! all

   ! useful global parameters
   real(double), parameter :: PI               = 3.141592653589793_double  
   real(double), parameter :: TWO_PI           = 2*PI      
   real(double), parameter :: HALF_PI          = PI/ 2
   real(double), parameter :: radiansToDegrees = 180/PI
   real(double), parameter :: degreesToRadians = PI/180

contains

   ! Info: Returns an sorted array
   ! Example:
   !     real(single)   myUnsortedArray(20)
   !     real(single)   mySortedArray(20)
   !     :
   !     :
   !     mySortedArray=sorted(myUnsortedArray)
   function sorted(unsortedArray) result(sortedArray)
      real(single),  intent(in)          :: unsortedArray(:)
      real(single)                       :: sortedArray(size(unsortedArray))

      sortedArray=unsortedArray(sortedIndexOf(unsortedArray))
   end function sorted



   ! Info: Returns an array of sorted indices.
   ! Example:
   !     real(single)   myArray(20)
   !     integer(int32) myIndex(20)
   !     :
   !     :
   !     myIndex=sortedIndexOf(myArray)
   function sortedIndexOf(array) result(sortedIndices)
      real(single),  intent(in)          :: array(:)
      integer(int32)                     :: sortedIndices(size(array))

      integer(int32)                     :: sizeArray
      integer(int32)                     :: i, ir, j, l, irra
      real(single)                       :: rra

      sizeArray=size(array)

      ! immediate return if:
      if (sizeArray == 0) return
      if (sizeArray < 2) then
         sortedIndices(1) = 1
         return
      end if

      ! initial
      do i = 1, sizeArray
         sortedIndices(i) = i
      end do

      l = sizeArray / 2 + 1
      ir = sizeArray   

      do 
         if (l > 1) then
            l = l - 1
            rra = array(sortedIndices(l))   
            irra = sortedIndices(l)    
         else
            rra = array(sortedIndices(ir))        
            irra = sortedIndices(ir)

            sortedIndices(ir) = sortedIndices(1)
            ir = ir - 1

            if (ir <= 1) then

               sortedIndices(1) = irra
               exit
            end if
         end if

         i = l
         j = l + l
         do while(j <= ir)

            if (j < ir) then
               if (array(sortedIndices(j)) < array(sortedIndices(j+1))) j=j+1
            end if
            if (rra < array(sortedIndices(j))) then

               sortedIndices(i) = sortedIndices(j)
               i = j
               j = j + j
            else
               exit
            end if

         end do

         sortedIndices(i) = irra
      end do
   end function sortedIndexOf



   ! Info: Returns the median of an unsorted array
   ! Example:
   !     real(single)   myArray(20)
   !     real(single)   median
   !     :
   !     :
   !     median=medianOfUnsorted(myArray)
   function medianOfUnsorted(array) result(median)
      real(single), intent(in) :: array(:)
      real(single)             :: median

      median=medianOfSorted(array(sortedIndexOf(array)))
   end function medianOfUnsorted



   ! Info: Returns the median of a sorted array
   ! Example:
   !     real(single)   myArray(20)
   !     real(single)   median
   !     :
   !     :
   !     median=medianOfSorted(myArray)
   function medianOfSorted(array) result(median)
      real(single), intent(in) :: array(:)
      real(single)             :: median

      integer(int32)           :: sizeArray
      integer(int32)           :: k


      sizeArray=size(array)

      ! immediate return if:
      if (sizeArray == 0) then 
         median=0.0
         return
      end if
      if (sizeArray == 1) then 
         median = array(1)
         return
      end if

      k = sizeArray / 2
      if (k * 2 .eq. sizeArray) then   ! Even number
         median  = (array(k) + array(k + 1)) / 2.0
      else
         median = array(k + 1)
      end if
   end function medianOfSorted



   ! Info: Calculates the area of a segment of a unity circle
   !       A segment is that portion of a disk whose upper boundary is a
   !       circular arc and whose lower boundary is a chord at distance d
   !       of the center.
   !       The chord must be perpendicular to 'd'.
   ! Limitations:
   !       0<=d<=1
   ! Result: A in range [0, 0.5]
   function areaOfCircleSegment(d) result (A)
      real(single),intent(in) :: d
      real(single)            :: A
      
      ! short cuts
      if (d==0.) then
         A=0.5
         return
      end if
      if (d==1.) then
         A=0.
         return
      end if

      A = ( acos(d) - d*sqrt(1 - d**2) ) / PI
   end function areaOfCircleSegment



   ! Info: Calculates the area 'D' in a circle with R=1
   !
   !       |----
   !       |    --
   !       |     |- -
   !       | A   |   -
   !       |     |  B  \
   !       |-----+----- \
   !       |     |       |
   !       | D   |  C    |
   !       |     |        |
   !       +--------------|
   ! Limitations: x must be in quadrant 1
   function areaInQuadrantUnder(x) result(A)
      real(single),intent(in) :: x(2)
      real(single)            :: A

      if ( sqrt(sum(x**2)) > 1. ) then
         A=HALF_PI * (.5 -areaOfCircleSegment(x(1)) -areaOfCircleSegment(x(2)))
      else
         A=x(1)*x(2)
      end if
   end function areaInQuadrantUnder



   ! Info: Calculates the fraction of a box with coordinates P1 and P2, that
   !       intersects a circle with origin (0,0) and R=1.
   ! Limitations: none?
   function fractionBoxInUnitCircle(p1,p2) result(f)
      real(single),intent(in) :: p1(2)
      real(single),intent(in) :: p2(2)
      real(single)            :: f

      real(single)            :: A,Abox
      real(single)            :: x1(2)
      real(single)            :: x2(2)
      logical                 :: sameDirection(2)
      real(single)            :: A11,A12,A21,A22

      if (any(p2-p1==0)) then
         f=0.
         return
      end if
      Abox=abs((p2(1)-p1(1))*(p2(2)-p1(2)))

      ! signs:
      sameDirection(1)=(p1(1)==0).or.(p2(1)==0).or.&
           sign(1.,p1(1))==sign(1.,p2(1))
      sameDirection(2)=(p1(2)==0).or.(p2(2)==0).or.&
           sign(1.,p1(2)) == sign(1.,p2(2))

      ! absolute points in mapped on Quadrant 0
      x1=min(abs(p1),abs(p2))
      x2=max(abs(p1),abs(p2))

      ! areas
      A11=areaInQuadrantUnder(x1)
      A12=areaInQuadrantUnder((/x1(1),x2(2)/))
      A21=areaInQuadrantUnder((/x2(1),x1(2)/))
      A22=areaInQuadrantUnder(x2)

      ! what method?
      if (sameDirection(1)) then
         if (sameDirection(2)) then
            ! same quadrant
            A=A22-(A12+A21-A11)
         else
            ! adjacent
            A=(A22-A21)+(A12-A11)
         end if

      else
         if (sameDirection(2)) then
            ! adjacent
            A=(A22-A21)+(A12-A11)
         else
            ! diagonal
            A=A11+A12+A21+A22
         end if
      end if

      f=A/Abox
   end function fractionBoxInUnitCircle



   ! Info: calculates the fraction of a box with coordinates P1 and P2, that
   !       intersects a circle with origin CE and and R=R
   ! Limitations: R>0
   function fractionBoxInCircle(ce,R,p1,p2) result(f)
      real(single),intent(in) :: ce(2)
      real(single),intent(in) :: R
      real(single),intent(in) :: p1(2)
      real(single),intent(in) :: p2(2)
      real(single)            :: f

      ! calculate the area within the box that was part of the circle
      f=fractionBoxInUnitCircle((p1-ce)/R,(p2-ce)/R)
   end function fractionBoxInCircle



   ! Info: Returns the quadrant of 2-dimensional coordinates!
   !       The result is in range [0, 4], where 0 denotes the coordinate
   !       is on axis...
   ! Limitations: none?
   function quadrantOfCoord(x)
      real(single),intent(in) :: x(2) ! 
      integer                 :: quadrantOfCoord

      if (all(x == 0.)) then
         quadrantOfCoord=0
         return
      end if
      if (x(1) > 0) then
         if (x(2) > 0) then
            quadrantOfCoord=1
         else
            quadrantOfCoord=2
         end if
      else
         if (x(2) > 0) then
            quadrantOfCoord=4
         else
            quadrantOfCoord=3
         end if
      end if
   end function quadrantOfCoord



   ! Info: Returns the quadrant of 2-dimensional coordinates!
   !       The result is in range [0, 4], where 0 denotes the coordinate
   !       is on axis...
   ! Example:
   !       call assertOn(associated(ptr),"Memory leak?")
   !       deallocate(ptr)
   !       nullify(ptr)
   subroutine assertOn(statement,comment)
      logical,     intent(in) :: statement
      character(*),intent(in) :: comment
      if (statement) then
         call fatal("assertOn",comment)
      end if
   end subroutine assertOn


end module omsourceutils_mod


